# Output backends for sigmac
# Copyright 2016-2018 Thomas Patzke

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sigma

### Mixins
class QuoteCharMixin:
    """
    This class adds the cleanValue method that quotes and filters characters according to the configuration in
    the attributes provided by the mixin.
    """
    reEscape = None                     # match characters that must be quoted
    escapeSubst = "\\\\\g<1>"           # Substitution that is applied to characters/strings matched for escaping by reEscape
    reClear = None                      # match characters that are cleaned out completely

    def cleanValue(self, val):
        if type(val) == int:
            return val
        if self.reEscape:
            val = self.reEscape.sub(self.escapeSubst, val)
        if self.reClear:
            val = self.reClear.sub("", val)
        return val

class RulenameCommentMixin:
    """Prefixes each rule with the rule title."""
    prefix = "# "
    options = (
            ("rulecomment", False, "Prefix generated query with comment containing title", None),
            )

    def generateBefore(self, parsed):
        if self.rulecomment:
            try:
                return "%s%s\n" % (self.prefix, parsed.sigmaParser.parsedyaml['title'])
            except KeyError:
                return ""

    def generateAfter(self, parsed):
        if self.rulecomment:
            return "\n"

class MultiRuleOutputMixin:
    """Mixin with common for multi-rule outputs"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.rulenames = set()

    def getRuleName(self, sigmaparser):
        """
        Generate a rule name from the title of the Sigma rule with following properties:

        * Spaces are replaced with -
        * Unique name by addition of a counter if generated name already in usage

        Generated names are tracked by the Mixin.

        """
        try:
            yaml_id = sigmaparser.parsedyaml["id"]
        except KeyError:
            yaml_id = "00000000-0000-0000-0000-000000000000"
        try:
            yaml_title = sigmaparser.parsedyaml["title"]
        except KeyError:
            yaml_title = "No Title"
        yaml_title = yaml_title.replace(" ", "-").replace("(", "").replace(")", "")
        
        rulename = yaml_id if yaml_id != "00000000-0000-0000-0000-000000000000" else yaml_title
        
        if rulename in self.rulenames:   # add counter if name collides
            cnt = 2
            while "%s-%d" % (rulename, cnt) in self.rulenames:
                cnt += 1
            rulename = "%s-%d" % (rulename, cnt)
        self.rulenames.add(rulename)

        return rulename



# Output backend discovery
# Copyright 2016-2019 Thomas Patzke, Florian Roth

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import json
import re
import os
import sigma.backends
from .base import BaseBackend
from sigma.tools import getAllSubclasses, getClassDict

def getBackendList():
    """Return list of backend classes"""
    path = os.path.dirname(__file__)
    return getAllSubclasses(path, "backends", BaseBackend)

def getBackendDict():
    return getClassDict(getBackendList())

def getBackend(name):
    try:
        return getBackendDict()[name]
    except KeyError as e:
        raise LookupError("Backend not found") from e



# Output backends for sigmac
# Copyright 2020 Jonas Hagg

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from sigma.backends.sql import SQLBackend
from sigma.parser.condition import NodeSubexpression, ConditionAND, ConditionOR, ConditionNOT
from sigma.parser.modifiers.type import SigmaRegularExpressionModifier
import re

class SQLiteBackend(SQLBackend):
    """Converts Sigma rule into SQL query for SQLite"""
    identifier = "sqlite"
    active = True

    # Single quoted string literals are preferred in SQlite 
    # https://www.sqlite.org/quirks.html#double_quoted_string_literals_are_accepted 
    valueExpression = "\'%s\'" # Expression of values, %s represents value
    mapFullTextSearch = "%s MATCH ('\"%s\"')"
    mapRegex = "%s REGEXP %s"
    typedValueExpression = { 
        SigmaRegularExpressionModifier: "\'%s\'" # Syntax for regular expressions
    }

    countFTS = 0

    def __init__(self, sigmaconfig, table):
        super().__init__(sigmaconfig, table)
        self.mappingItem = False

    def requireFTS(self, node):
        return (not self.mappingItem and
                (type(node) in (int, str) or all(isinstance(val, str) for val in node) or all(isinstance(val, int) for val in node)))

    def generateFTS(self, value):
        if re.search(r"((\\(\*|\?|\\))|\*|\?|_|%)", value):
            raise NotImplementedError(
                "Wildcards in SQlite Full Text Search not implemented")
        self.countFTS += 1
        return self.mapFullTextSearch % (self.table, value)

    def generateANDNode(self, node):

        if self.requireFTS(node):
            fts = str('"' + self.andToken + '"').join(self.cleanValue(val)
                                                      for val in node)
            return self.generateFTS(fts)

        generated = [self.generateNode(val) for val in node]
        filtered = [g for g in generated if g is not None]
        if filtered:
            return self.andToken.join(filtered)
        else:
            return None

    def generateORNode(self, node):

        if self.requireFTS(node):
            fts = str('"' + self.orToken + '"').join(self.cleanValue(val)
                                                     for val in node)
            return self.generateFTS(fts)

        generated = [self.generateNode(val) for val in node]
        filtered = [g for g in generated if g is not None]
        if filtered:
            return self.orToken.join(filtered)
        else:
            return None
            
    def cleanValue(self, val):
        if not isinstance(val, str):
            return str(val)

        # Escape single quotes in SQLite
        val = val.replace('\'','\'\'')

        # Single backlashes which are not in front of * or ? are doulbed
        val = re.sub(r"(?<!\\)\\(?!(\\|\*|\?))", r"\\\\", val)

        # Replace _ with \_ because _ is a sql wildcard
        val = re.sub(r'_', r'\_', val)

        # Replace % with \% because % is a sql wildcard
        val = re.sub(r'%', r'\%', val)

        # Replace * with %, if even number of backslashes (or zero) in front of *
        val = re.sub(r"(?<!\\)(\\\\)*(?!\\)\*", r"\1%", val)

        # Replace ? with _, if even number of backsashes (or zero) in front of ?
        val = re.sub(r"(?<!\\)(\\\\)*(?!\\)\?", r"\1_", val)

        return val

    def generateMapItemNode(self, node):
        try:
            self.mappingItem = True
            fieldname, value = node
            transformed_fieldname = self.fieldNameMapping(fieldname, value)

            if value is None: value = '' # Handle null values

            generated_value = self.generateNode(value)

            has_wildcard = re.search(
                r"((\\(\*|\?|\\))|\*|\?|_|%)", generated_value) 

            if "," in generated_value and generated_value[0]=="(" and generated_value[-1]==")" and not has_wildcard:
                return self.mapMulti % (transformed_fieldname, generated_value)
            elif type(value) == SigmaRegularExpressionModifier:
                return self.mapRegex % (transformed_fieldname, generated_value) # regex are mapped "as is"
            elif "LENGTH" in transformed_fieldname:
                return self.mapLength % (transformed_fieldname, value)
            elif type(value) == list:
                return self.generateMapItemListNode(transformed_fieldname, value)
            elif self.mapListsSpecialHandling == False and type(value) in (str, int, list) or self.mapListsSpecialHandling == True and type(value) in (str, int):
                if has_wildcard:
                    return self.mapWildcard % (transformed_fieldname, generated_value)
                else:
                    return self.mapExpression % (transformed_fieldname, generated_value)
            elif "sourcetype" in transformed_fieldname:
                return self.mapSource % (transformed_fieldname, generated_value)
            elif has_wildcard:
                return self.mapWildcard % (transformed_fieldname, generated_value)
            else:
                raise TypeError(
                    "Backend does not support map values of type " + str(type(value)))
        finally:
            self.mappingItem = False

    def generateValueNode(self, node):
        if self.mappingItem:
            return self.valueExpression % (self.cleanValue(str(node)))
        else:
            return self.generateFTS(self.cleanValue(str(node)))

    def generateTypedValueNode(self, node):
        return self.typedValueExpression[type(node)] % (str(node))

    def generateQuery(self, parsed):
        self.countFTS = 0
        return self._generateQueryWithFields(parsed, list("*"))

    def checkFTS(self, parsed, result):
        if self.countFTS > 1:
            raise NotImplementedError(
                "Match operator ({}) is allowed only once in SQLite, parse rule in a different way:\n{}".format(self.countFTS, result))
        self.countFTS = 0



# Sigma parser
# Copyright 2016-2017 Thomas Patzke, Florian Roth

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
from .base import SimpleParser
from .exceptions import SigmaParseError

COND_NONE = 0
COND_AND  = 1
COND_OR   = 2
COND_NOT  = 3
COND_NULL = 4


# Debugging code
def dumpNode(node, indent=''):   # pragma: no cover
    """
    Recursively print the AST rooted at *node* for debugging.
    """
    if hasattr(node, 'items'):
        print("%s%s<%s>" % (indent, type(node).__name__,
                            type(node.items).__name__))
        if type(node.items) != list:
            dumpNode(node.items, indent + '  ')
        else:
            for item in node.items:
                dumpNode(item, indent + '  ')
    else:
        print("%s%s=%s" % (indent, type(node).__name__,
                                   repr(node)))
    return node


# Condition Tokenizer
class SigmaConditionToken:
    """Token of a Sigma condition expression"""
    TOKEN_AND    = 1
    TOKEN_OR     = 2
    TOKEN_NOT    = 3
    TOKEN_ID     = 4
    TOKEN_LPAR   = 5
    TOKEN_RPAR   = 6
    TOKEN_PIPE   = 7
    TOKEN_ONE    = 8
    TOKEN_ALL    = 9
    TOKEN_AGG    = 10
    TOKEN_EQ     = 11
    TOKEN_LT     = 12
    TOKEN_LTE    = 13
    TOKEN_GT     = 14
    TOKEN_GTE    = 15
    TOKEN_BY     = 16
    TOKEN_NEAR   = 17

    tokenstr = [
            "INVALID",
            "AND",
            "OR",
            "NOT",
            "ID",
            "LPAR",
            "RPAR",
            "PIPE",
            "ONE",
            "ALL",
            "AGG",
            "EQ",
            "LT",
            "LTE",
            "GT",
            "GTE",
            "BY",
            "NEAR",
            ]

    def __init__(self, tokendef, match, pos):
        self.type = tokendef[0]
        self.matched = match.group()
        self.pos = pos

    def __eq__(self, other):
        if type(other) == int:      # match against type
            return self.type == other
        if type(other) == str:      # match against content
            return self.matched == other
        else:
            raise NotImplementedError("SigmaConditionToken can only be compared against token type constants")

    def __str__(self):  # pragma: no cover
        return "[ Token: %s: '%s' ]" % (self.tokenstr[self.type], self.matched)


class SigmaConditionTokenizer:
    """Tokenize condition string into token sequence"""
    tokendefs = [      # list of tokens, preferred recognition in given order, (token identifier, matching regular expression). Ignored if token id == None
            (SigmaConditionToken.TOKEN_ONE,    re.compile("1 of", re.IGNORECASE)),
            (SigmaConditionToken.TOKEN_ALL,    re.compile("all of", re.IGNORECASE)),
            (None,       re.compile("[\\s\\r\\n]+")),
            (SigmaConditionToken.TOKEN_AGG,    re.compile("count|min|max|avg|sum", re.IGNORECASE)),
            (SigmaConditionToken.TOKEN_NEAR,   re.compile("near", re.IGNORECASE)),
            (SigmaConditionToken.TOKEN_BY,     re.compile("by", re.IGNORECASE)),
            (SigmaConditionToken.TOKEN_EQ,     re.compile("==")),
            (SigmaConditionToken.TOKEN_LTE,    re.compile("<=")),
            (SigmaConditionToken.TOKEN_LT,     re.compile("<")),
            (SigmaConditionToken.TOKEN_GTE,    re.compile(">=")),
            (SigmaConditionToken.TOKEN_GT,     re.compile(">")),
            (SigmaConditionToken.TOKEN_PIPE,   re.compile("\\|")),
            (SigmaConditionToken.TOKEN_AND,    re.compile("and", re.IGNORECASE)),
            (SigmaConditionToken.TOKEN_OR,     re.compile("or", re.IGNORECASE)),
            (SigmaConditionToken.TOKEN_NOT,    re.compile("not", re.IGNORECASE)),
            (SigmaConditionToken.TOKEN_ID,     re.compile("[\\w*-.]+")),
            (SigmaConditionToken.TOKEN_LPAR,   re.compile("\\(")),
            (SigmaConditionToken.TOKEN_RPAR,   re.compile("\\)")),
            ]

    def __init__(self, condition):
        if type(condition) == str:          # String that is parsed
            self.tokens = list()
            pos = 1

            while len(condition) > 0:
                for tokendef in self.tokendefs:     # iterate over defined tokens and try to recognize the next one
                    match = tokendef[1].match(condition)
                    if match:
                        if tokendef[0] != None:
                            self.tokens.append(SigmaConditionToken(tokendef, match, pos + match.start()))
                        pos += match.end()      # increase position and cut matched prefix from condition
                        condition = condition[match.end():]
                        break
                else:   # no valid token identified
                    raise SigmaParseError("Unexpected token in condition at position %s" % condition)
        elif type(condition) == list:       # List of tokens to be converted into SigmaConditionTokenizer class
            self.tokens = condition
        else:
            raise TypeError("SigmaConditionTokenizer constructor expects string or list, got %s" % (type(condition)))

    def __str__(self):  # pragma: no cover
        return " ".join([str(token) for token in self.tokens])

    def __iter__(self):
        return iter(self.tokens)

    def __len__(self):
        return len(self.tokens)

    def __getitem__(self, i):
        if type(i) == int:
            return self.tokens[i]
        elif type(i) == slice:
            return SigmaConditionTokenizer(self.tokens[i])
        else:
            raise IndexError("Expected index or slice")

    def __add__(self, other):
        if isinstance(other, SigmaConditionTokenizer):
            return SigmaConditionTokenizer(self.tokens + other.tokens)
        elif isinstance(other, (SigmaConditionToken, ParseTreeNode)):
            return SigmaConditionTokenizer(self.tokens + [ other ])
        else:
            raise TypeError("+ operator expects SigmaConditionTokenizer or token type, got %s: %s" % (type(other), str(other)))

    def index(self, item):
        return self.tokens.index(item)


### Parse Tree Node Classes ###
class ParseTreeNode:
    """Parse Tree Node Base Class"""
    def __init__(self):
        raise NotImplementedError("ConditionBase is no usable class")

    def __str__(self):  # pragma: no cover
        return "[ %s: %s ]" % (self.__doc__, str([str(item) for item in self.items]))


class ConditionBase(ParseTreeNode):
    """Base class for conditional operations"""
    op = COND_NONE
    items = None

    def __init__(self, sigma=None, op=None, *args):
        if type(self) == ConditionBase:
            raise NotImplementedError("ConditionBase is no usable class")

        if sigma == None and op == None and len(args) == 0:    # no parameters given - initialize empty
            self.items = list()
        else:       # called by parser, use given values
            self.items = args

    def add(self, item):
        self.items.append(item)

    def __iter__(self):
        return iter(self.items)

    def __len__(self):
        return len(self.items)


class ConditionBaseOneItem(ConditionBase):
    def __init__(self, sigma=None, op=None, val=None):
        if type(self) == ConditionBaseOneItem:
            raise NotImplementedError("ConditionBaseOneItem is no usable class")

        if sigma == None and op == None and val == None:    # no parameters given - initialize empty
            self.items = list()
        else:       # called by parser, use given values
            self.items = [ val ]

    def add(self, item):
        if len(self.items) == 0:
            super.add(item)
        else:
            raise ValueError("Only one element allowed")

    @property
    def item(self):
        try:
            return self.items[0]
        except IndexError:
            return None


class ConditionAND(ConditionBase):
    """AND Condition"""
    op = COND_AND


class ConditionOR(ConditionBase):
    """OR Condition"""
    op = COND_OR


class ConditionNOT(ConditionBaseOneItem):
    """NOT Condition"""
    op = COND_NOT


class ConditionNULLValue(ConditionBaseOneItem):
    """Condition: Field value is empty or doesn't exists"""
    op = COND_NULL
    pass


class ConditionNotNULLValue(ConditionNULLValue):
    """Condition: Field value is not empty"""
    op = COND_NULL
    pass


class NodeSubexpression(ParseTreeNode):
    """Subexpression"""
    def __init__(self, subexpr):
        self.items = subexpr


class SigmaSearchValueAsIs:
    """The contained value is used as-is in the output."""
    def __init__(self, value):
        self.value = value


# Parse tree generators: generate parse tree nodes from extended conditions
def generateXOf(sigma, val, condclass):
    """
    Generic implementation of (1|all) of x expressions.

    * condclass across all list items if x is name of definition
    * condclass across all definitions if x is keyword 'them'
    * condclass across all matching definition if x is wildcard expression, e.g. 'selection*'
    """
    if val.matched == "them":           # OR across all definitions
        cond = condclass()
        for name, definition in sigma.definitions.items():
            if name == "timeframe":
                continue
            cond.add(NodeSubexpression(sigma.parse_definition(definition)))
        return NodeSubexpression(cond)
    elif val.matched.find("*") > 0:     # OR across all matching definitions
        cond = condclass()
        reDefPat = re.compile("^" + val.matched.replace("*", ".*") + "$")
        for name, definition in sigma.definitions.items():
            if name != "timeframe" and reDefPat.match(name):
                cond.add(NodeSubexpression(sigma.parse_definition(definition)))
        return NodeSubexpression(cond)
    else:                               # OR across all items of definition
        return NodeSubexpression(sigma.parse_definition_byname(val.matched, condclass))


def generateAllOf(sigma, op, val):
    """Convert 'all of x' expressions into ConditionAND"""
    return generateXOf(sigma, val, ConditionAND)


def generateOneOf(sigma, op, val):
    """Convert '1 of x' expressions into ConditionOR"""
    return generateXOf(sigma, val, ConditionOR)


def convertId(sigma, op):
    """Convert search identifiers (lists or maps) into condition nodes according to spec defaults"""
    return NodeSubexpression(sigma.parse_definition_byname(op.matched))


# Optimizer
class SigmaConditionOptimizer:
    """
    Optimizer for the parsed AST.
    """
    def _stripSubexpressionNode(self, node):
        """
        Recursively strips all subexpressions (i.e. brackets) from the AST.
        """
        if type(node) == NodeSubexpression:
            assert(type(node.items) != list)
            return self._stripSubexpressionNode(node.items)
        if hasattr(node, 'items') and type(node) is not ConditionNOT:
            node.items = list(map(self._stripSubexpressionNode, node.items))
        return node

    def _unstripSubexpressionNode(self, node):
        """
        Recursively adds brackets around AND and OR operations in the AST.
        """
        if type(node) in (ConditionAND, ConditionOR):
            newnode = NodeSubexpression(node)
            node.items = list(map(self._unstripSubexpressionNode, node.items))
            return newnode
        return node

    def _ordered_uniq(self, l):
        """
        Remove duplicate entries in list *l* while preserving order.

        Used to be fast before it needed to work around list instead of
        tuple being used for lists within definitions in the AST.
        """
        seen = set()
        #return [x for x in l if x not in seen and not seen.add(x)]
        uniq = []
        for x in l:
            if type(x) == tuple and type(x[1]) == list:
                x = (x[0], tuple(x[1]))
            if x not in seen and not seen.add(x):
                uniq.append(x)
        out = []
        for x in uniq:
            if type(x) == tuple and type(x[1]) == tuple:
                out.append((x[0], list(x[1])))
            else:
                out.append(x)
        return out

    def _optimizeNode(self, node, changes=False):
        """
        Recursively optimize the AST rooted at *node* once.  Returns the new
        root node and a boolean indicating if the tree was changed in this
        invocation or any of the recursive sub-invocations.

        You MUST remove all subexpression nodes from the AST before calling
        this function.  Subexpressions are implicit around AND/OR nodes.
        """
        if type(node) in (ConditionOR, ConditionAND):
            # Remove empty OR(X), AND(X)
            if len(node.items) == 0:
                return None, True
            if None in node.items:
                node.items = [item for item in node.items if item != None]
                return self._optimizeNode(node, changes=True)

            # OR(X), AND(X)                 =>  X
            if len(node.items) == 1:
                return self._optimizeNode(node.items[0], changes=True)

            # OR(X, X, ...), AND(X, X, ...) =>  OR(X, ...), AND(X, ...)
            uniq_items = self._ordered_uniq(node.items)
            if len(uniq_items) < len(node.items):
                node.items = uniq_items
                return self._optimizeNode(node, changes=True)

            # OR(X, OR(Y))                  =>  OR(X, Y)
            if any(type(child) == type(node) for child in node.items) and \
               all(type(child) in (type(node), tuple) for child in node.items):
                newitems = []
                for child in node.items:
                    if hasattr(child, 'items'):
                        newitems.extend(child.items)
                    else:
                        newitems.append(child)
                node.items = newitems
                return self._optimizeNode(node, changes=True)

            # fallthrough

        elif type(node) == ConditionNOT:
            assert(len(node.items) == 1)
            # NOT(NOT(X))                   =>  X
            if type(node.items[0]) == ConditionNOT:
                assert(len(node.items[0].items) == 1)
                return self._optimizeNode(node.items[0].items[0], changes=True)

            # NOT(ConditionNULLValue)       =>  ConditionNotNULLValue
            if type(node.items[0]) == ConditionNULLValue:
                newnode = ConditionNotNULLValue(val=node.items[0].items[0])
                return self._optimizeNode(newnode, changes=True)

            # NOT(ConditionNotNULLValue)    =>  ConditionNULLValue
            if type(node.items[0]) == ConditionNotNULLValue:
                newnode = ConditionNULLValue(val=node.items[0].items[0])
                return self._optimizeNode(newnode, changes=True)

            # fallthrough

        else:
            return node, changes

        itemresults = [self._optimizeNode(item, changes) for item in node.items]
        node.items = [res[0] for res in itemresults]
        if any(res[1] for res in itemresults):
            changes = True
        return node, changes

    def optimizeTree(self, tree):
        """
        Optimize the boolean expressions in the AST rooted at *tree*.

        The main idea behind optimizing the AST is that less repeated terms is
        generally better for backend performance.  This is especially relevant
        to backends that do not perform any query language optimization down
        the road, such as those that generate code.

        A common example for when these suboptimal rules actually occur in
        practice is when a rule has multiple alternative detections that are
        OR'ed together in the condition, and all of the detections include a
        common element, such as the same EventID.

        The following optimizations are currently performed:
        -   Removal of empty OR(), AND()
        -   OR(X), AND(X)                 =>  X
        -   OR(X, X, ...), AND(X, X, ...) =>  OR(X, ...), AND(X, ...)
        -   OR(X, OR(Y))                  =>  OR(X, Y)
        -   OR(AND(X, ...), AND(X, ...))  =>  AND(X, OR(AND(...), AND(...)))
        -   NOT(NOT(X))                   =>  X
        -   NOT(ConditionNULLValue)       =>  ConditionNotNULLValue
        -   NOT(ConditionNotNULLValue)    =>  ConditionNULLValue

        Boolean logic simplification is NP-hard.  To avoid backtracking,
        speculative transformations that may or may not lead to a more optimal
        expression were not implemented.  These include for example factoring
        out common operands that are not in all, but only some AND()s within an
        OR(), or vice versa.  Nevertheless, it is safe to assume that this
        implementation performs poorly on very large expressions.
        """
        tree = self._stripSubexpressionNode(tree)
        changes = True
        while changes:
            tree, changes = self._optimizeNode(tree)
        tree = self._unstripSubexpressionNode(tree)
        return tree

# Condition parser
class SigmaConditionParser:
    """Parser for Sigma condition expression"""
    searchOperators = [     # description of operators: (token id, number of operands, parse tree node class) - order == precedence
            (SigmaConditionToken.TOKEN_ALL, 1, generateAllOf),
            (SigmaConditionToken.TOKEN_ONE, 1, generateOneOf),
            (SigmaConditionToken.TOKEN_ID,  0, convertId),
            (SigmaConditionToken.TOKEN_NOT, 1, ConditionNOT),
            (SigmaConditionToken.TOKEN_AND, 2, ConditionAND),
            (SigmaConditionToken.TOKEN_OR,  2, ConditionOR),
            ]

    def __init__(self, sigmaParser, tokens):
        self.sigmaParser = sigmaParser
        self.config = sigmaParser.config
        self._optimizer = SigmaConditionOptimizer()

        if SigmaConditionToken.TOKEN_PIPE in tokens:    # Condition contains atr least one aggregation expression
            pipepos = tokens.index(SigmaConditionToken.TOKEN_PIPE)
            self.parsedSearch = self.parseSearch(tokens[:pipepos])
            self.parsedAgg = SigmaAggregationParser(tokens[pipepos + 1:], self.sigmaParser, self.config)
        else:
            self.parsedSearch = self.parseSearch(tokens)
            self.parsedAgg = None

    def parseSearch(self, tokens, depth=0):
        """
        Iterative parsing of search expression.
        """
        def find_close_token_index_in_pairs(tokens, start_index, open_token, close_token):
            """
                the function try to find close_token index for open_token in pairs
                e.g
                    open_token was '(' and
                    tokens were ['(', '...', '(', '...', ')', ')']
                    the first '(' should pair with the last ')' instead of the first ')'

                Parameters:
                    tokens: the list of tokens
                    start_index: the start index (included) of the input tokens for finding the close_token
                    open_token: the token that considered as opening token
                    close_token: the token that considered as closing token
                Returns:
                    the index of the close_token in pair with the open_token
                    raise ValueError when there is no close_token in pairs
            """
            open_token_count = 0
            for i in range(start_index, len(tokens)):
                if tokens[i] == open_token:
                    open_token_count += 1
                elif tokens[i] == close_token:
                    if open_token_count == 0:
                        return i
                    else:
                        open_token_count -= 1
            raise ValueError(f"matched close_token {close_token} is not found in tokens")
        # 1. Identify subexpressions with parentheses around them and parse them like a separate search expression
        while SigmaConditionToken.TOKEN_LPAR in tokens:
            lPos = tokens.index(SigmaConditionToken.TOKEN_LPAR)
            lTok = tokens[lPos]
            try:
                rPos = find_close_token_index_in_pairs(tokens, lPos+1, SigmaConditionToken.TOKEN_LPAR, SigmaConditionToken.TOKEN_RPAR)
                rTok = tokens[rPos]
            except ValueError as e:
                raise SigmaParseError("Missing matching closing parentheses") from e
            if lPos + 1 == rPos:
                raise SigmaParseError("Empty subexpression at " + str(lTok.pos))
            if lPos > rPos:
                raise SigmaParseError("Closing parentheses at position " + str(rTok.pos) + " precedes opening at position " + str(lTok.pos))

            subparsed = self.parseSearch(tokens[lPos + 1:rPos], depth=depth+1)
            tokens = tokens[:lPos] + NodeSubexpression(subparsed) + tokens[rPos + 1:]   # replace parentheses + expression with group node that contains parsed subexpression

        # 2. Iterate over all known operators in given precedence
        for operator in self.searchOperators:
            # 3. reduce all occurrences into corresponding parse tree nodes
            while operator[0] in tokens:
                pos_op = tokens.index(operator[0])
                tok_op = tokens[pos_op]
                if operator[1] == 0:    # operator
                    treenode = operator[2](self.sigmaParser, tok_op)
                    tokens = tokens[:pos_op] + treenode + tokens[pos_op + 1:]
                elif operator[1] == 1:    # operator value
                    pos_val = pos_op + 1
                    tok_val = tokens[pos_val]
                    treenode = operator[2](self.sigmaParser, tok_op, tok_val)
                    tokens = tokens[:pos_op] + treenode + tokens[pos_val + 1:]
                elif operator[1] == 2:    # value1 operator value2
                    pos_val1 = pos_op - 1
                    pos_val2 = pos_op + 1
                    tok_val1 = tokens[pos_val1]
                    tok_val2 = tokens[pos_val2]
                    treenode = operator[2](self.sigmaParser, tok_op, tok_val1, tok_val2)
                    tokens = tokens[:pos_val1] + treenode + tokens[pos_val2 + 1:]

        if len(tokens) != 1:     # parse tree must begin with exactly one node
            raise ValueError("Parse tree must have exactly one start node!")
        query_cond = tokens[0]

        # 4. Integrate conditions from logsources in configurations to outermost expression
        if depth == 0:
            ls_cond = self.sigmaParser.get_logsource_condition()
            if ls_cond is not None:
                cond = ConditionAND()
                cond.add(ls_cond)
                cond.add(query_cond)
                query_cond = cond

        return self._optimizer.optimizeTree(query_cond)

    def __str__(self):  # pragma: no cover
        return str(self.parsedSearch)

    def __len__(self):  # pragma: no cover
        return len(self.parsedSearch)


# Aggregation parser
class SigmaAggregationParser(SimpleParser):
    """Parse Sigma aggregation expression and provide parsed data"""
    parsingrules = [
            {   # State 0
                SigmaConditionToken.TOKEN_AGG:  ("aggfunc", "trans_aggfunc", 1),
                SigmaConditionToken.TOKEN_NEAR: ("aggfunc", "init_near_parsing", 8),
            },
            {   # State 1
                SigmaConditionToken.TOKEN_LPAR: (None, None, 2)
            },
            {   # State 2
                SigmaConditionToken.TOKEN_RPAR: (None, None, 4),
                SigmaConditionToken.TOKEN_ID: ("aggfield", "trans_fieldname", 3),
            },
            {   # State 3
                SigmaConditionToken.TOKEN_RPAR: (None, None, 4)
            },
            {   # State 4
                SigmaConditionToken.TOKEN_BY: ("cond_op", None, 5),
                SigmaConditionToken.TOKEN_EQ: ("cond_op", None, 7),
                SigmaConditionToken.TOKEN_LT: ("cond_op", None, 7),
                SigmaConditionToken.TOKEN_LTE: ("cond_op", None, 7),
                SigmaConditionToken.TOKEN_GT: ("cond_op", None, 7),
                SigmaConditionToken.TOKEN_GTE: ("cond_op", None, 7),
            },
            {   # State 5
                SigmaConditionToken.TOKEN_ID: ("groupfield", "trans_fieldname", 6)
            },
            {   # State 6
                SigmaConditionToken.TOKEN_EQ: ("cond_op", None, 7),
                SigmaConditionToken.TOKEN_LT: ("cond_op", None, 7),
                SigmaConditionToken.TOKEN_LTE: ("cond_op", None, 7),
                SigmaConditionToken.TOKEN_GT: ("cond_op", None, 7),
                SigmaConditionToken.TOKEN_GTE: ("cond_op", None, 7),
            },
            {   # State 7
                SigmaConditionToken.TOKEN_ID: ("condition", None, -1)
            },
            {   # State 8
                SigmaConditionToken.TOKEN_ID: (None, "store_search_id", 9)
            },
            {   # State 9
                SigmaConditionToken.TOKEN_AND: (None, "set_include", 10),
            },
            {   # State 10
                SigmaConditionToken.TOKEN_NOT: (None, "set_exclude", 8),
                SigmaConditionToken.TOKEN_ID: (None, "store_search_id", 9),
            },
            ]
    finalstates = {-1, 9}

    # Aggregation functions
    AGGFUNC_COUNT = 1
    AGGFUNC_MIN   = 2
    AGGFUNC_MAX   = 3
    AGGFUNC_AVG   = 4
    AGGFUNC_SUM   = 5
    AGGFUNC_NEAR  = 6
    aggfuncmap = {
            "count": AGGFUNC_COUNT,
            "min":   AGGFUNC_MIN,
            "max":   AGGFUNC_MAX,
            "avg":   AGGFUNC_AVG,
            "sum":   AGGFUNC_SUM,
            "near":  AGGFUNC_NEAR,
            }

    def __init__(self, tokens, parser, config):
        self.parser = parser
        self.config = config
        self.aggfield = None
        self.groupfield = None
        super().__init__(tokens)

    def trans_aggfunc(self, name):
        """Translate aggregation function name into constant"""
        try:
            return self.aggfuncmap[name]
        except KeyError:
            raise SigmaParseError("Unknown aggregation function '%s'" % (name))

    def trans_fieldname(self, fieldname):
        """Translate field name into configured mapped name"""
        mapped = self.config.get_fieldmapping(fieldname).resolve_fieldname(fieldname, self.parser)
        if type(mapped) == str:
            return mapped
        else:
            raise NotImplementedError("Field mappings in aggregations must be single valued")

    def init_near_parsing(self, name):
        """Initialize data structures for 'near" aggregation operator parsing"""
        self.include = list()
        self.exclude = list()
        self.current = self.include
        return self.trans_aggfunc(name)

    def store_search_id(self, name):
        self.current.append(name)
        return name

    def set_include(self, name):
        self.current = self.include

    def set_exclude(self, name):
        self.current = self.exclude



# Sigma parser
# Copyright 2016-2018 Thomas Patzke, Florian Roth

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Rule Filtering
import datetime
class SigmaRuleFilter:
    """Filter for Sigma rules with conditions"""
    LEVELS = {
            "low"      : 0,
            "medium"   : 1,
            "high"     : 2,
            "critical" : 3
            }
    STATES = [
              "unsupported",
              "deprecated",
              "experimental",
              "test",
              "stable"]

    def __init__(self, expr):
        self.minlevel      = None
        self.maxlevel      = None
        self.status        = None
        self.notstatus     = None
        self.tlp           = None
        self.target        = None
        self.logsources    = list()
        self.notlogsources = list()
        self.tags          = list()
        self.nottags       = list()
        self.inlastday     = None
        self.condition     = list()
        self.notcondition  = list()

        for cond in [c.replace(" ", "") for c in expr.split(",")]:
            if cond.startswith("level<="):
                try:
                    level = cond[cond.index("=") + 1:]
                    self.maxlevel = self.LEVELS[level]
                except KeyError as e:
                    raise SigmaRuleFilterParseException("Unknown level '%s' in condition '%s'" % (level, cond)) from e
            elif cond.startswith("level>="):
                try:
                    level = cond[cond.index("=") + 1:]
                    self.minlevel = self.LEVELS[level]
                except KeyError as e:
                    raise SigmaRuleFilterParseException("Unknown level '%s' in condition '%s'" % (level, cond)) from e
            elif cond.startswith("level="):
                try:
                    level = cond[cond.index("=") + 1:]
                    self.minlevel = self.LEVELS[level]
                    self.maxlevel = self.minlevel
                except KeyError as e:
                    raise SigmaRuleFilterParseException("Unknown level '%s' in condition '%s'" % (level, cond)) from e
            elif cond.startswith("status="):
                self.status = cond[cond.index("=") + 1:]
                if self.status not in self.STATES:
                    raise SigmaRuleFilterParseException("Unknown status '%s' in condition '%s'" % (self.status, cond))
            elif cond.startswith("status!="):
                self.notstatus = cond[cond.index("=") + 1:]
                if self.notstatus not in self.STATES:
                    raise SigmaRuleFilterParseException("Unknown status '%s' in condition '%s'" % (self.notstatus, cond))
            elif cond.startswith("tlp="):
                self.tlp = cond[cond.index("=") + 1:].upper()  #tlp is always uppercase
            elif cond.startswith("target="):
                self.target = cond[cond.index("=") + 1:].lower() # lower to make caseinsensitive
            elif cond.startswith("logsource="):
                self.logsources.append(cond[cond.index("=") + 1:])
            elif cond.startswith("logsource!="):
                self.notlogsources.append(cond[cond.index("=") + 1:])
            elif cond.startswith("tag="):
                self.tags.append(cond[cond.index("=") + 1:].lower())
            elif cond.startswith("tag!="):
                self.nottags.append(cond[cond.index("=") + 1:].lower())
            elif cond.startswith("condition="):
                self.condition.append(cond[cond.index("=") + 1:].lower())
            elif cond.startswith("condition!="):
                self.notcondition.append(cond[cond.index("=") + 1:].lower())
            elif cond.startswith("inlastday="):
                nbday = cond[cond.index("=") + 1:]
                try:
                    self.inlastday = int(nbday)
                except ValueError as e:
                    raise SigmaRuleFilterParseException("Unknown number '%s' in condition '%s'" % (nbday, cond)) from e
            else:
                raise SigmaRuleFilterParseException("Unknown condition '%s'" % cond)

    def match(self, yamldoc):
        """Match filter conditions against rule"""
        # Levels
        if self.minlevel is not None or self.maxlevel is not None:
            try:
                level = self.LEVELS[yamldoc['level']]
            except KeyError:    # missing or invalid level
                return False    # User wants level restriction, but it's not possible here

            # Minimum level
            if self.minlevel is not None:
                if level < self.minlevel:
                    return False
            # Maximum level
            if self.maxlevel is not None:
                if level > self.maxlevel:
                    return False

        # Status
        if self.status is not None:
            try:
                status = yamldoc['status']
            except KeyError:    # missing status
                return False    # User wants status restriction, but it's not possible here
            if status != self.status:
                return False

        if self.notstatus is not None:
            try:
                status = yamldoc['status']
            except KeyError:    # missing status
                return False    # User wants status restriction, but it's not possible here
            if status == self.notstatus:
                return False


        # Tlp
        if self.tlp is not None:
            try:
                tlp = yamldoc['tlp']
            except KeyError:    # missing tlp
                tlp = "WHITE"    # tlp is WHITE by default
            if tlp != self.tlp:
                return False

        #Target
        if self.target:
            try:
                targets = [ target.lower() for target in yamldoc['target']]
            except (KeyError, AttributeError):    # no target set
                return False
            if self.target not in targets:
                return False

        # Log Sources
        if self.logsources:
            try:
                logsources = { value for key, value in yamldoc['logsource'].items() }
            except (KeyError, AttributeError):    # no log source set
                return False    # User wants status restriction, but it's not possible here

            for logsrc in self.logsources:
                if logsrc not in logsources:
                    return False

        # NOT Log Sources
        if self.notlogsources:
            try:
                notlogsources = { value for key, value in yamldoc['logsource'].items() }
            except (KeyError, AttributeError):    # no log source set
                return False    # User wants status restriction, but it's not possible here

            for logsrc in self.notlogsources:
                if logsrc  in notlogsources:
                    return False

        # Tags
        if self.tags:
            try:
                tags = [ tag.lower() for tag in yamldoc['tags']]
            except (KeyError, AttributeError):    # no tags set
                return False

            for tag in self.tags:
                if tag not in tags:
                    return False
        # NOT Tags
        if self.nottags:
            try:
                nottags = [ tag.lower() for tag in yamldoc['tags']]
            except (KeyError, AttributeError):    # no tags set
                return False

            for tag in self.nottags:
                if tag in nottags:
                    return False

        # date in the last N days
        if self.inlastday:
           try:
               date_str = yamldoc['date']
           except KeyError:    # missing date
               return False    # User wants date time restriction, but it's not possible here

           try:
               modified_str = yamldoc['modified']
           except KeyError:    # no update
               modified_str = None
           if modified_str:
               date_str = modified_str

           date_object = datetime.datetime.strptime(date_str, '%Y/%m/%d')
           today_objet = datetime.datetime.now()
           delta       = today_objet - date_object
           if delta.days > self.inlastday:
                return False

        if self.condition:
            try:
                conditions = yamldoc['detection']['condition']
                if isinstance(conditions,list):                         # sone time conditions are list even with only 1 line
                    s_condition = ' '.join(conditions)
                else:
                    s_condition = conditions
            except KeyError:    # missing condition
                return False    # User wants condition restriction, but it's not possible here
            for val in self.condition:
                if not val in s_condition:
                    return False

        if self.notcondition:
            try:
                conditions = yamldoc['detection']['condition']
                if isinstance(conditions,list):                         # sone time conditions are list even with only 1 line
                    s_condition = ' '.join(conditions)
                else:
                    s_condition = conditions
            except KeyError:    # missing condition
                return False    # User wants condition restriction, but it's not possible here
            for val in self.notcondition:
                if val in s_condition:
                    return False

        # all tests passed
        return True

class SigmaRuleFilterParseException(Exception):
    pass



# Code used across all Sigma modules
# Copyright 2016-2019 Thomas Patzke, Florian Roth

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pkgutil
import importlib

def getAllSubclasses(path, import_base, base_class):
    """Return list of all classes derived from a superclass contained in a module."""
    classes = set()
    for finder, name, ispkg in pkgutil.iter_modules([ path ]):
        module = importlib.import_module(".{}.{}".format(import_base, name), __package__)
        for name, cls in vars(module).items():
            if type(cls) == type and issubclass(cls, base_class) and cls.active:
                classes.add(cls)
    return classes

def getClassDict(clss):
    """Return a dictionary: class.identifier -> class"""
    return {cls.identifier: cls for cls in clss }



#!/usr/bin/env python3
# Assign UUIDs to Sigma rules and verify UUID assignment for a Sigma rule repository

from argparse import ArgumentParser
from pathlib import Path
from uuid import uuid4, UUID
import yaml
from sigma.output import SigmaYAMLDumper

argparser = ArgumentParser(description="Assign and verify UUIDs of Sigma rules")
argparser.add_argument("--verify", "-V", action="store_true", help="Verify existence and uniqueness of UUID assignments. Exits with error code if verification fails.")
argparser.add_argument("--verbose", "-v", action="store_true", help="Be verbose.")
argparser.add_argument("--recursive", "-r", action="store_true", help="Recurse into directories.")
argparser.add_argument("--error", "-e", action="store_true", help="Exit with error code 10 on verification failures.")
argparser.add_argument("inputs", nargs="+", help="Sigma rule files or repository directories")
args = argparser.parse_args()

if args.recursive:
    paths = [ p for pathname in args.inputs for p in Path(pathname).glob("**/*") if p.is_file() ]
else:
    paths = [ Path(pathname) for pathname in args.inputs ]

def print_verbose(*arg, **kwarg):
    if args.verbose:
        print(*arg, **kwarg)

# Define order-preserving representer from dicts/maps
def yaml_preserve_order(self, dict_data):
    return self.represent_mapping("tag:yaml.org,2002:map", dict_data.items())

yaml.add_representer(dict, yaml_preserve_order)

uuids = set()
passed = True
for path in paths:
    print_verbose("Rule {}".format(str(path)))
    with path.open("r") as f:
        rules = list(yaml.safe_load_all(f))

    if args.verify:
        i = 1
        for rule in rules:
            if "title" in rule:     # Rule with a title should also have a UUID
                try:
                    UUID(rule["id"])
                except ValueError:  # id is not a valid UUID
                    print("Rule {} in file {} has a malformed UUID '{}'.".format(i, str(path), rule["id"]))
                    passed = False
                except KeyError:    # rule has no id
                    print("Rule {} in file {} has no UUID.".format(i, str(path)))
                    passed = False
            i += 1
    else:
        newrules = list()
        changed = False
        i = 1
        for rule in rules:
            if "title" in rule and "id" not in rule:    # only assign id to rules that have a title and no id
                newrule = dict()
                changed = True
                for k, v in rule.items():
                    newrule[k] = v
                    if k == "title":    # insert id after title
                        uuid = uuid4()
                        newrule["id"] = str(uuid)
                        print("Assigned UUID '{}' to rule {} in file {}.".format(uuid, i, str(path)))
                newrules.append(newrule)
            else:
                newrules.append(rule)
            i += 1

        if changed:
            with path.open("w") as f:
                yaml.dump_all(newrules, f, Dumper=SigmaYAMLDumper, indent=4, width=160, default_flow_style=False)

if not passed:
    print("The Sigma rules listed above don't have an ID. The ID must be:")
    print("* Contained in the 'id' attribute")
    print("* a valid UUIDv4 (randomly generated)")
    print("* Unique in this repository")
    print("Please generate one with the sigma-uuid tool or here: https://www.uuidgenerator.net/version4")
    if args.error:
        exit(10)



#!/usr/bin/env python3
# A Sigma to SIEM converter
# Copyright 2016-2017 Thomas Patzke, Florian Roth
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import argparse
import yaml
import ruamel.yaml
import json
import pathlib
import itertools
import logging, traceback
from sigma.parser.collection import SigmaCollectionParser
from sigma.parser.exceptions import SigmaCollectionParseError, SigmaParseError
from sigma.configuration import SigmaConfiguration, SigmaConfigurationChain
from sigma.config.collection import SigmaConfigurationManager
from sigma.config.exceptions import SigmaConfigParseError, SigmaRuleFilterParseException
from sigma.filter import SigmaRuleFilter
import sigma.backends.discovery as backends
from sigma.backends.base import BackendOptions
from sigma.backends.exceptions import BackendError, NotSupportedError, PartialMatchError, FullMatchError
from sigma.parser.modifiers import modifiers
import codecs
import copy
import time
import datetime
from termcolor import colored

sys.stdout = codecs.getwriter('utf-8')(sys.stdout.detach())

# Error codes

ERR_OUTPUT              = 1
ERR_INVALID_YAML        = 3
ERR_SIGMA_PARSING       = 4
ERR_OPEN_SIGMA_RULE     = 5
ERR_OPEN_CONFIG_FILE    = 5
ERR_CONFIG_INVALID_YAML = 6
ERR_CONFIG_PARSING      = 6
ERR_BACKEND             = 8
ERR_NOT_SUPPORTED       = 9
ERR_NO_TARGET           = 10
ERR_RULE_FILTER_PARSING = 11
ERR_CONFIG_REQUIRED     = 20
ERR_CONFIG_ORDER        = 21
ERR_CONFIG_BACKEND      = 22
ERR_OUTPUT_FORMAT       = 30
ERR_NOT_IMPLEMENTED     = 42
ERR_PARTIAL_FIELD_MATCH = 80
ERR_FULL_FIELD_MATCH    = 90

# Allowed fields in output
allowed_fields = ["title", "id", "status", "description", "author", "references", "fields", "falsepositives", "level", "tags", "filename"]

deprecation_warning_message = colored("Sigmac will be deprecated by the end of 2022",
                                      "red") + " in favour of sigma-cli and pySigma. Please " +  colored("stop contributing backends", "red") + \
                                               " to this tool. Limited support is offered until the end of 2023, " \
                                               "especially for backends that haven't been migrated yet.\n "


def alliter(path):
    for sub in path.iterdir():
        if sub.name.startswith("."):
            continue
        if sub.is_dir():
            yield from alliter(sub)
        else:
            yield sub

def get_inputs(paths, recursive):
    if paths == ['-']:
        return [sys.stdin]

    if recursive:
        return list(itertools.chain.from_iterable([list(alliter(pathlib.Path(p))) for p in paths]))
    else:
        return [pathlib.Path(p) for p in paths]

class ActionBackendHelp(argparse.Action):
    def __call__(self, parser, ns, vals, opt):
        backend = backends.getBackend(vals)
        if len(backend.options) > 0:
            helptext = "Backend options for " + backend.identifier + "\n"
            for option, default, help, _ in backend.options:
                helptext += "    {:10}: {} (default: {})".format(option, help, default) + "\n"

        print(helptext)
        exit(0)

def set_argparser():
    """Sets up and parses the command line arguments for Sigmac.
    Returns the argparser"""
    argparser = argparse.ArgumentParser(description="Convert Sigma rules into SIEM signatures.\n" + deprecation_warning_message, formatter_class=argparse.RawTextHelpFormatter)
    argparser.add_argument("--recurse", "-r", action="store_true", help="Use directory as input (recurse into subdirectories is not implemented yet)")
    argparser.add_argument("--filter", "-f", help="""
    Define comma-separated filters that must match (AND-linked) to rule to be processed.
    Valid filters: level<=x, level>=x, level=x, status=y, logsource=z, tag=t, target=o.
    x is one of: low, medium, high, critical.
    y is one of: experimental, testing, stable.
    z is a word appearing in an arbitrary log source attribute.
    t is a tag that must appear in the rules tag list, case-insensitive matching.
    o is a target that must appear in the rules target list, case-insensitive matching.
    Multiple log source specifications are AND linked.
    Special filter:
    inlastday=X rule create or modified in the last X days period
    tlp=valid_tlp if rule have no tlp set to WHITE
            """)
    argparser.add_argument("--target", "-t", choices=backends.getBackendDict().keys(), help="Output target format")
    argparser.add_argument("--lists", "-l", action="store_true", help="List available output target formats and configurations")
    argparser.add_argument("--lists-files-after-date", "-L",help="List yml files  which is modified/created after the date (Example of the date: 2022/02/01).")
    argparser.add_argument("--config", "-c", action="append", help="Configurations with field name and index mapping for target environment. Multiple configurations are merged into one. Last config is authoritative in case of conflicts.")
    argparser.add_argument("--output", "-o", default=None, help="Output file or filename prefix (if end with a '_','/' or '\\')")
    argparser.add_argument("--output-fields", "-of", help="""Enhance your output with additional fields from the Sigma rule (not only the converted rule itself).
    Select the fields you want by providing their list delimited with commas (no space). Only work with the '--output-format' option and with 'json' or 'yaml' value.
    available additional fields : title, id, status, description, author, references, fields, falsepositives, level, tags.
    This option do not have any effect for backends that already format output : elastalert, kibana, splukxml etc. """)
    argparser.add_argument("--output-format", "-oF", choices=["json", "yaml"], help="Use only if you want to have JSON or YAML output (default is raw text)")
    argparser.add_argument("--output-extention", "-e", default=None, help="Extension of Output file for filename prefix use")
    argparser.add_argument("--print0", action="store_true", help="Delimit results by NUL-character")
    argparser.add_argument("--backend-option", "-O", action="append", help="Options and switches that are passed to the backend")
    argparser.add_argument("--backend-config", "-C", help="Configuration file (YAML format) containing options to pass to the backend")
    argparser.add_argument("--backend-help", action=ActionBackendHelp, help="Print backend options")
    argparser.add_argument("--defer-abort", "-d", action="store_true", help="Don't abort on parse or conversion errors, proceed with next rule. The exit code from the last error is returned")
    argparser.add_argument("--ignore-backend-errors", "-I", action="store_true", help="Only return error codes for parse errors and ignore errors for rules that cause backend errors. Useful, when you want to get as much queries as possible.")
    argparser.add_argument("--shoot-yourself-in-the-foot", action="store_true", help=argparse.SUPPRESS)
    argparser.add_argument("--verbose", "-v", action="store_true", help="Be verbose")
    argparser.add_argument("--debug", "-D", action="store_true", help="Debugging output")
    argparser.add_argument("inputs", nargs="*", help="Sigma input files ('-' for stdin)")

    return argparser

def list_backends(debug):
    for backend in sorted(backends.getBackendList(), key=lambda backend: backend.identifier):
        if debug:
            print("{:>15} : {} ({})".format(backend.identifier, backend.__doc__, backend.__name__))
        else:
            print("{:>15} : {}".format(backend.identifier, backend.__doc__))

def list_configurations(backend=None, scm=None):
    for conf_id, title, backends in sorted(scm.list(), key=lambda config: config[0]):
        if backend is not None and backend in backends or backend is None or len(backends) == 0:
            print("{:>30} : {}".format(conf_id, title))

def list_modifiers(modifiers):
    for modifier_id, modifier in modifiers.items():
        print("{:>10} : {}".format(modifier_id, modifier.__doc__))

def get_fileName_after_date(inputs, recurse, date):
    #date: 2021/05/06
    #modified: 2021/11/30
    dateTime = time.mktime(datetime.datetime.strptime(date, "%Y/%m/%d").timetuple())
    for sigmafile in get_inputs(inputs, recurse):
        f = sigmafile.open(encoding='utf-8')
        yamls = yaml.safe_load_all(f)
        datestr = None
        modifiedstr = None
        for data in yamls:
            modifiedstr = data.get("modified", None)
            datestr = data.get("date", None)
            if not modifiedstr and not datestr:
                continue;

        if not modifiedstr and not datestr:
            print("%s, No date" % sigmafile)
            continue

        if not modifiedstr:
            modifiedstr = datestr

        modified = time.mktime(datetime.datetime.strptime(modifiedstr, "%Y/%m/%d").timetuple())
        if modified > dateTime:
            print("%s, Updated" % sigmafile)

def main():
    argparser = set_argparser()
    cmdargs = argparser.parse_args()
    scm = SigmaConfigurationManager()

    logger = logging.getLogger(__name__)
    if cmdargs.debug:   # pragma: no cover
        logging.basicConfig(filename='sigmac.log', filemode='w', level=logging.DEBUG)
        logger.setLevel(logging.DEBUG)

    if cmdargs.lists:
        print("Backends (Targets):")
        list_backends(cmdargs.debug)

        print()
        print("Configurations (Sources):")
        list_configurations(backend=cmdargs.target, scm=scm)

        print()
        print("Modifiers:")
        list_modifiers(modifiers=modifiers)
        sys.exit(0)
    elif len(cmdargs.inputs) == 0:
        print("Nothing to do!")
        print(deprecation_warning_message)
        argparser.print_usage()
        sys.exit(0)

    if cmdargs.lists_files_after_date is not None:
        get_fileName_after_date(cmdargs.inputs, cmdargs.recurse, cmdargs.lists_files_after_date)
        sys.exit(0)

    if cmdargs.target is None:
        print("No target selected, select one with -t/--target")
        argparser.print_usage()
        sys.exit(ERR_NO_TARGET)

    logger.debug("* Target selected %s" % (cmdargs.target))

    rulefilter = None
    if cmdargs.filter:
        try:
            rulefilter = SigmaRuleFilter(cmdargs.filter)
        except SigmaRuleFilterParseException as e:
            print("Parse error in Sigma rule filter expression: %s" % str(e), file=sys.stderr)
            sys.exit(ERR_RULE_FILTER_PARSING)

    sigmaconfigs = SigmaConfigurationChain()
    backend_class = backends.getBackend(cmdargs.target)
    if cmdargs.config is None:
        if backend_class.config_required and not cmdargs.shoot_yourself_in_the_foot:
            print("The backend you want to use usually requires a configuration to generate valid results. Please provide one with --config/-c.", file=sys.stderr)
            print("Available choices for this backend (get complete list with --lists/-l):")
            list_configurations(backend=cmdargs.target, scm=scm)
            sys.exit(ERR_CONFIG_REQUIRED)
        if backend_class.default_config is not None:
            cmdargs.config = backend_class.default_config

    if cmdargs.config:
        order = 0
        for conf_name in cmdargs.config:
            try:
                sigmaconfig = scm.get(conf_name)
                if sigmaconfig.order is not None:
                    if sigmaconfig.order <= order and not cmdargs.shoot_yourself_in_the_foot:
                        print("The configurations were provided in the wrong order (order key check in config file)", file=sys.stderr)
                        sys.exit(ERR_CONFIG_ORDER)
                    order = sigmaconfig.order

                try:
                    if cmdargs.target not in sigmaconfig.config["backends"]:
                        print("The configuration '{}' is not valid for backend '{}'. Valid choices are: {}".format(conf_name, cmdargs.target, ", ".join(sigmaconfig.config["backends"])), file=sys.stderr)
                        sys.exit(ERR_CONFIG_ORDER)
                except KeyError:
                    pass

                sigmaconfigs.append(sigmaconfig)
            except OSError as e:
                print("Failed to open Sigma configuration file %s: %s" % (conf_name, str(e)), file=sys.stderr)
                exit(ERR_OPEN_CONFIG_FILE)
            except (yaml.parser.ParserError, yaml.scanner.ScannerError) as e:
                print("Sigma configuration file %s is no valid YAML: %s" % (conf_name, str(e)), file=sys.stderr)
                exit(ERR_CONFIG_INVALID_YAML)
            except SigmaConfigParseError as e:
                print("Sigma configuration parse error in %s: %s" % (conf_name, str(e)), file=sys.stderr)
                exit(ERR_CONFIG_PARSING)

    if cmdargs.output_fields:
        if cmdargs.output_format:
            output_fields_rejected = [field for field in cmdargs.output_fields.split(",") if field not in allowed_fields] # Not allowed fields
            if output_fields_rejected:
                    print("These fields are not allowed (check help for allow field list) : %s" % (", ".join(output_fields_rejected)), file=sys.stderr)
                    exit(ERR_OUTPUT_FORMAT)
            else:
                output_fields_filtered = [field for field in cmdargs.output_fields.split(",") if field in allowed_fields] # Keep only allowed fields
        else:
            print("The '--output-fields' or '-of' arguments must be used with '--output-format' or '-oF' equal to 'json' or 'yaml'", file=sys.stderr)
            exit(ERR_OUTPUT_FORMAT)

    backend_options = BackendOptions(cmdargs.backend_option, cmdargs.backend_config)
    backend = backend_class(sigmaconfigs, backend_options)

    filename_ext = cmdargs.output_extention
    filename = cmdargs.output
    fileprefix = None
    if filename:
        if filename_ext:
            if filename_ext[0] == '.':
                pass
            else:
                filename_ext = '.' + filename_ext
        else:
            filename_ext = '.rule'

        if filename[-1:] in ['_','/','\\']:
            fileprefix = filename
        else:
            try:
                out = open(filename, "w", encoding='utf-8')
            except (IOError, OSError) as e:
                print("Failed to open output file '%s': %s" % (filename, str(e)), file=sys.stderr)
                exit(ERR_OUTPUT)
    else:
        out = sys.stdout

    error = 0
    output_array = []
    result = backend.initialize()
    if result:
        print(result, file=out)

    for sigmafile in get_inputs(cmdargs.inputs, cmdargs.recurse):
        logger.debug("* Processing Sigma input %s" % (sigmafile))
        success = True
        try:
            if cmdargs.inputs == ['-']:
                f = sigmafile
            else:
                f = sigmafile.open(encoding='utf-8')
            parser = SigmaCollectionParser(f, sigmaconfigs, rulefilter, sigmafile)
            backend.setYmlFileName(str(sigmafile))
            results = parser.generate(backend)

            nb_result = len(list(copy.deepcopy(results)))
            inc_filenane = None if nb_result < 2 else 0

            newline_separator = '\0' if cmdargs.print0 else '\n'

            results = list(results) # Since results is an iterator and used twice we convert it a list
            for result in results:
                if not fileprefix == None and not inc_filenane == None: #yml action
                    try:
                        filename = fileprefix + str(sigmafile.name)
                        filename = filename.replace('.yml','_' + str(inc_filenane) + filename_ext)
                        inc_filenane += 1
                        out = open(filename, "w", encoding='utf-8')
                    except (IOError, OSError) as e:
                        print("Failed to open output file '%s': %s" % (filename, str(e)), file=sys.stderr)
                        exit(ERR_OUTPUT)
                elif  not fileprefix == None and inc_filenane == None: # a simple yml
                    try:
                        filename = fileprefix + str(sigmafile.name)
                        filename = filename.replace('.yml',filename_ext)
                        out = open(filename, "w", encoding='utf-8')
                    except (IOError, OSError) as e:
                        print("Failed to open output file '%s': %s" % (filename, str(e)), file=sys.stderr)
                        exit(ERR_OUTPUT)
                if not cmdargs.output_fields:
                    print(result, file=out, end=newline_separator)

            if cmdargs.output_fields: # Handle output fields
                output={}
                f.seek(0)
                docs = yaml.load_all(f, Loader=yaml.FullLoader)
                for doc in docs:
                    for k,v in doc.items():
                        if k in output_fields_filtered:
                            output[k] = v
                output['rule'] = [result for result in results]
                if len(output['rule']) > 0: # avoid printing empty rules
                    if "filename" in output_fields_filtered:
                        output['filename'] = str(sigmafile.name)
                    output_array.append(output)

            if nb_result == 0: # backend get only 1 output
                if not fileprefix == None: # want a prefix anyway
                    try:
                        filename = "%s%s_mono_output%s" % (fileprefix,cmdargs.target,filename_ext)
                        out = open(filename, "w", encoding='utf-8')
                        fileprefix = None  # no need to open the same file many time
                    except (IOError, OSError) as e:
                        print("Failed to open output file '%s': %s" % (filename, str(e)), file=sys.stderr)
                        exit(ERR_OUTPUT)

        except OSError as e:
            print("Failed to open Sigma file %s: %s" % (sigmafile, str(e)), file=sys.stderr)
            logger.debug("* Convertion Sigma input %s FAILURE" % (sigmafile))
            success = False
            error = ERR_OPEN_SIGMA_RULE
        except (yaml.parser.ParserError, yaml.scanner.ScannerError) as e:
            print("Error: Sigma file %s is no valid YAML: %s" % (sigmafile, str(e)), file=sys.stderr)
            logger.debug("* Convertion Sigma input %s FAILURE" % (sigmafile))
            success = False
            error = ERR_INVALID_YAML
            if not cmdargs.defer_abort:
                sys.exit(error)
        except (SigmaParseError, SigmaCollectionParseError) as e:
            print("Error: Sigma parse error in %s: %s" % (sigmafile, str(e)), file=sys.stderr)
            logger.debug("* Convertion Sigma input %s FAILURE" % (sigmafile))
            success = False
            error = ERR_SIGMA_PARSING
            if not cmdargs.defer_abort:
                sys.exit(error)
        except NotSupportedError as e:
            print("Error: The Sigma rule requires a feature that is not supported by the target system: " + str(e), file=sys.stderr)
            logger.debug("* Convertion Sigma input %s FAILURE" % (sigmafile))
            success = False
            if not cmdargs.ignore_backend_errors:
                error = ERR_NOT_SUPPORTED
                if not cmdargs.defer_abort:
                    sys.exit(error)
        except BackendError as e:
            print("Error: Backend error in %s: %s" % (sigmafile, str(e)), file=sys.stderr)
            logger.debug("* Convertion Sigma input %s FAILURE" % (sigmafile))
            success = False
            if not cmdargs.ignore_backend_errors:
                error = ERR_BACKEND
                if not cmdargs.defer_abort:
                    sys.exit(error)
        except (NotImplementedError, TypeError) as e:
            print("An unsupported feature is required for this Sigma rule (%s): " % (sigmafile) + str(e), file=sys.stderr)
            # traceback.print_exc()
            logger.debug("* Convertion Sigma input %s FAILURE" % (sigmafile))
            success = False
            if not cmdargs.ignore_backend_errors:
                error = ERR_NOT_IMPLEMENTED
                if not cmdargs.defer_abort:
                    sys.exit(error)
        except PartialMatchError as e:
            print("Error: Partial field match error: %s" % str(e), file=sys.stderr)
            logger.debug("* Convertion Sigma input %s FAILURE" % (sigmafile))
            success = False
            if not cmdargs.ignore_backend_errors:
                error = ERR_PARTIAL_FIELD_MATCH
                if not cmdargs.defer_abort:
                    sys.exit(error)
        except FullMatchError as e:
            print("Error: Full field match error", file=sys.stderr)
            logger.debug("* Convertion Sigma input %s FAILURE" % (sigmafile))
            success = False
            if not cmdargs.ignore_backend_errors:
                error = ERR_FULL_FIELD_MATCH
                if not cmdargs.defer_abort:
                    sys.exit(error)
        finally:
            try:
                f.close()
            except:
                pass

        if success :
            logger.debug("* Convertion Sigma input %s SUCCESS" % (sigmafile))

    result = backend.finalize()
    if result:
        print(result, file=out)

    if cmdargs.output_fields:
        if cmdargs.output_format == 'json':
            print(json.dumps(output_array, indent=4, ensure_ascii=False), file=out)
        elif cmdargs.output_format == 'yaml':
            print(ruamel.yaml.round_trip_dump(output_array), file=out)

    out.close()

    sys.exit(error)

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Import given Sigma rules to MISP

import argparse
import pathlib
import urllib3
urllib3.disable_warnings()
from pymisp import PyMISP, MISPEvent

def create_new_event(args, misp):
    if hasattr(misp, "new_event"):
        return misp.new_event(info=args.info)["Event"]["id"]
    
    event = MISPEvent()
    event.info = args.info
    return misp.add_event(event)["Event"]["id"]


class MISPImportArgumentParser(argparse.ArgumentParser):
    def __init__(self, *args, **kwargs):
        super().__init__(
            description="Import Sigma rules into MISP events",
            epilog="Parameters can be read from a file by a @filename parameter. The file should contain one parameter per line. Dashes may be omitted.",
            fromfile_prefix_chars="@",
        )

    def convert_arg_line_to_args(self, line : str):
        return ("--" + line.lstrip("--")).split()

def main():
    argparser = MISPImportArgumentParser()
    argparser.add_argument("--url", "-u", default="https://localhost", help="URL of MISP instance")
    argparser.add_argument("--key", "-k", required=True, help="API key")
    argparser.add_argument("--insecure", "-I", action="store_false", help="Disable TLS certificate validation.")
    argparser.add_argument("--event", "-e", type=int, help="Add Sigma rule to event with this ID. If not set, create new event.")
    argparser.add_argument("--same-event", "-s", action="store_true", help="Import all Sigma rules to the same event, if no event is set.")
    argparser.add_argument("--info", "-i", default="Sigma import", help="Event Information field for newly created MISP event.")
    argparser.add_argument("--recursive", "-r", action="store_true", help="Recursive traversal of directory")
    argparser.add_argument("sigma", nargs="+", help="Sigma rule file that should be imported")
    args = argparser.parse_args()

    if args.recursive:
        paths = [ p for pathname in args.sigma for p in pathlib.Path(pathname).glob("**/*") if p.is_file() ]
    else:
        paths = [ pathlib.Path(sigma) for sigma in args.sigma ]

    misp = PyMISP(args.url, args.key, args.insecure)
    if args.event:
        if hasattr(misp, "get"):
            eventid = misp.get(args.event)["Event"]["id"]
        else:
            eventid = misp.get_event(args.event)["Event"]["id"]

    first = True

    for sigma in paths:
        if not args.event and (first or not args.same_event):
            eventid = create_new_event(args, misp)
        print("Importing Sigma rule {} into MISP event {}...".format(sigma, eventid, end=""))
        f = sigma.open("rt")

        if hasattr(misp, "add_named_attribute"):
            misp.add_named_attribute(eventid, "sigma", f.read())
        else:
            event = misp.get_event(eventid, pythonify=True)
            event.add_attribute("sigma", f.read())
            misp.update_event(event)

        f.close()
        first = False

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Convert Sigma rules with EventIDs to rules with generic log sources

from argparse import ArgumentParser
import yaml
import sys
from pathlib import Path
from sigma.output import SigmaYAMLDumper

class Output(object):
    """Output base class"""
    def write(self, *args, **kwargs):
        self.f.write(*args, **kwargs)

class SingleFileOutput(Output):
    """Output into single file with multiple YAML documents. Each input file is announced with comment."""
    def __init__(self, name):
        self.f = open(name, "x")
        self.path = None
        self.first = True

    def new_output(self, path):
        """Announce new Sigma rule as input and start new YAML document."""
        if self.path is None or self.path != path:
            if self.first:
                self.first = False
            else:
                self.f.write("---\n")
            self.path = path
            self.f.write("# Sigma rule: {}\n".format(path))

    def finish(self):
        self.f.close()

class StdoutOutput(SingleFileOutput):
    """Like SingleFileOutput, just for standard output"""
    def __init__(self):
        self.f = sys.stdout
        self.path = None
        self.first = True

    def finish(self):
        pass

class DirectoryOutput(Output):
    """Output each input file into a corresponding output file in target directory."""
    def __init__(self, dirpath):
        self.d = dirpath
        self.f = None
        self.path = None
        self.opened = None

    def new_output(self, path):
        if self.path is None or self.path != path:
            if self.f is not None:
                self.f.close()
            self.path = path
            self.opened = False     # opening file is deferred to first write

    def write(self, *args, **kwargs):
        if not self.opened:
            self.f = (self.d / self.path.name).open("x")
        super().write(*args, **kwargs)

    def finish(self):
        if self.f is not None:
            self.f.close()

def get_output(output):
    if output is None:
        return StdoutOutput()

    path = Path(output)
    if path.is_dir():
        return DirectoryOutput(path)
    else:
        return SingleFileOutput(output)

class AmbiguousRuleException(TypeError):
    def __init__(self, ids):
        super().__init__()
        self.ids = ids

    def __str__(self):
        return(", ".join([str(eid) for eid in self.ids]))

def convert_to_generic(yamldoc):
    changed = False
    try:
        product = yamldoc["logsource"]["product"]
        service = yamldoc["logsource"]["service"]
    except KeyError:
        return False

    if product == "windows" and service in ("sysmon", "security"):
        # Currently, only Windows Security or Sysmon are relevant
        eventids = set()
        for name, detection in yamldoc["detection"].items():      # first collect all event ids
            if name == "condition" or type(detection) is not dict:
                continue

            try:
                eventid = detection["EventID"]
                try:    # expect that EventID attribute contains a list
                    eventids.update(eventid)
                except TypeError:   # if this fails, it's a plain value
                    eventids.add(eventid)
            except KeyError:    # No EventID attribute
                pass

        if 1 in eventids and service == "sysmon" or \
                4688 in eventids and service == "security":
            if len(eventids) == 1:      # only convert if one EventID collected, else it gets complicated
                # remove all EventID definitions
                empty_name = list()
                for name, detection in yamldoc["detection"].items():
                    if name == "condition" or type(detection) is not dict:
                        continue
                    try:
                        del detection["EventID"]
                    except KeyError:
                        pass

                    if detection == {}:     # detection was reduced to nothing - remove it later
                        empty_name.append(name)

                for name in empty_name:     # delete empty detections
                    del yamldoc["detection"][name]

                if yamldoc["detection"] == {}:  # delete detection section if empty
                    del yamldoc["detection"]

                # rewrite log source
                yamldoc["logsource"] = {
                        "category": "process_creation",
                        "product":  "windows"
                        }

                changed = True
            else:                       # raise an exception to print a warning message to make user aware about the issue
                raise AmbiguousRuleException(eventids)
    return changed

def get_input_paths(args):
    if args.recursive:
        return [ p for pathname in args.sigma for p in Path(pathname).glob("**/*") if p.is_file() ]
    else:
        return [ Path(sigma) for sigma in args.sigma ]

argparser = ArgumentParser(description="Convert between classical and generic log source Sigma rules.")
argparser.add_argument("--output", "-o", help="Output file or directory. Default: standard output.")
argparser.add_argument("--recursive", "-r", action="store_true", help="Recursive traversal of directory")
argparser.add_argument("--converted-list", "-c", help="Write list of rule files that were successfully converted (default: stdout)")
argparser.add_argument("sigma", nargs="+", help="Sigma rule file(s) that should be converted")
args = argparser.parse_args()

# Define order-preserving representer from dicts/maps
def yaml_preserve_order(self, dict_data):
    return self.represent_mapping("tag:yaml.org,2002:map", dict_data.items())

yaml.add_representer(dict, yaml_preserve_order)

input_paths = get_input_paths(args)
output = get_output(args.output)
if args.converted_list:
    fconv = open(args.converted_list, "w")
else:
    fconv = sys.stdout

for path in input_paths:
    try:
        f = path.open("r")
    except OSError as e:
        print("Error while reading Sigma rule {}: {}".format(path, str(e)), file=sys.stderr)
        sys.exit(1)

    try:
        yamldocs = list(yaml.safe_load_all(f))
    except yaml.YAMLError as e:
        print("YAML parse error while parsing Sigma rule {}: {}".format(path, str(e)), file=sys.stderr)
        sys.exit(2)

    yamldoc_num = 0
    changed = False
    for yamldoc in yamldocs:
        yamldoc_num += 1
        output.new_output(path)
        try:
            changed |= convert_to_generic(yamldoc)
        except AmbiguousRuleException as e:
            changed = False
            print("Rule {} in file {} contains multiple EventIDs: {}".format(yamldoc_num, str(path), str(e)), file=sys.stderr)

    yamldocs_idx = list(zip(range(len(yamldocs)), yamldocs))
    delete = set()
    for i, yamldoc_a in yamldocs_idx:     # iterate over all yaml document pairs
        for j, yamldoc_b in yamldocs_idx:
            if j <= i:      # symmetric relation, skip same comparisons
                continue
            if yamldoc_a == yamldoc_b:
                delete.add(j)

    for i in reversed(sorted(delete)):  # delete double yaml documents
        del yamldocs[i]

    # Common special case: two yaml docs, one global and one remainder of multiple following docs - merge them
    try:
        if len(yamldocs) == 2 and \
                yamldocs[0]["action"] == "global" and \
                "action" not in yamldocs[1] and \
                set(yamldocs[0].keys()) & set(yamldocs[1].keys()) == set():   # last condition: no common keys
            yamldocs[0].update(yamldocs[1])
            del yamldocs[1]
    except KeyError:
        pass

    if changed:     # only write output if changed
        try:
            output.write(yaml.dump_all(yamldocs, Dumper=SigmaYAMLDumper, indent=4, width=160, default_flow_style=False))
            print(path, file=fconv)
        except OSError as e:
            print("Error while writing result: {}".format(str(e)), file=sys.stderr)
            sys.exit(2)

output.finish()



# Output backends for sigmac
# Copyright 2016-2017 Thomas Patzke, Ben de Haan

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
import sigma
from .base import SingleTextQueryBackend

class LogPointBackend(SingleTextQueryBackend):
    """Converts Sigma rule into LogPoint query"""
    identifier = "logpoint"
    active = True
    config_required = False
    default_config = ["sysmon", "logpoint-windows"]

    # \   -> \\
    # \*  -> \*
    # \\* -> \\*
    reEscape = re.compile('("|(?<!\\\\)\\\\(?![*?\\\\]))')
    reClear = None
    andToken = " "
    orToken = " OR "
    notToken = " -"
    subExpression = "(%s)"
    listExpression = "[%s]"
    listSeparator = ", "
    valueExpression = "\"%s\""
    nullExpression = "-%s=*"
    notNullExpression = "%s=*"
    mapExpression = "%s=%s"
    mapListsSpecialHandling = True
    mapListValueExpression = "%s IN %s"

    def generateAggregation(self, agg):
        if agg == None:
            return ""
        if agg.aggfunc == sigma.parser.condition.SigmaAggregationParser.AGGFUNC_NEAR:
            raise NotImplementedError("The 'near' aggregation operator is not yet implemented for this backend")
        if agg.groupfield == None:
            return " | chart %s(%s) as val | search val %s %s" % (agg.aggfunc_notrans, agg.aggfield or "", agg.cond_op, agg.condition)
        else:
            return " | chart %s(%s) as val by %s | search val %s %s" % (agg.aggfunc_notrans, agg.aggfield or "", agg.groupfield or "", agg.cond_op, agg.condition)



# Output backends for sigmac
# Copyright 2016-2018 Thomas Patzke, Florian Roth

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import csv 
import sigma
import yaml
import re
import copy

from sigma.backends.exceptions import NotSupportedError
from .mixins import RulenameCommentMixin, QuoteCharMixin
from sigma.parser.modifiers.base import SigmaTypeModifier
from .base import BaseBackend

class FortisemBackend(RulenameCommentMixin, BaseBackend, QuoteCharMixin):
    """Base class for Fortisem backends that generate one text-based expression from a Sigma rule"""
    identifier = "fortisiem"
    active = True

    reEscape = re.compile('("|(?<!\\\\)\\\\(?![*?\\\\]))')
    reClear = None
    andToken = " AND "
    orToken = " OR "
    notToken = "NOT "
    subExpression = "(%s)"
    listExpression = "(%s)"
    listSeparator = " "
    valueExpression = "%s"
    strValueExpression = "\"%s\""
    nullExpression = "%s IS NULL"
    notNullExpression = "%s IS NOT NULL"
    mapExpression = "%s=%s"
    mapExpressionNot = "%s!=%s"
    mapListsSpecialHandling = True
    mapListValueExpression = "%s IN (%s)"
    regExpression = "%s REGEXP ( %s )"
    mapListValueExpressionNot = "%s NOT IN (%s)"
    regExpressionNot = "%s NOT REGEXP ( %s )"

    sort_condition_lists = False        # Sort condition items for AND and OR conditions

    ymlAttr2FortiSIEMAttr = {}
    fortiSIEMAttrType = {}
    fileFilterDicts= {}
    WindowsSysmonCode2FortiSIEMEvtTy = {}

    product = None
    service = None
    category = None
    curAttrs= set()
    sourceValueForWinAppEvtTy= None

    #if Attribute value is too long, this rule will be skip.
    isValTooLong = False
    MAX_LEN = 2271

    techniqueMap = {}
    ruleType = None
    ruleIndex = 1;

    def __init__(self, sigmaconfig, backend_options=dict()):
        """
        Initialize backend. This gets a sigmaconfig object, which is notified about the used backend class by
        passing the object instance to it.
        """
        super().__init__(sigmaconfig, backend_options)
        self.setRuleType(backend_options)
        self.loadCSVfiles()
        self.loadMitreAttackMatrixFile(backend_options);

    def initialize(self):
        return "<Rules>"

    def finalize(self):
        return "</Rules>"
     
    # It's used to check whether the format of yml file is right. 
    def ymlValidator(self, node,regdicts={}):
        if type(node) == sigma.parser.condition.ConditionAND:
            for val in node:
                if self.ymlValidator(val) == False:
                    return False
            return True
        elif type(node) == sigma.parser.condition.ConditionOR:
            for val in node:
                if self.ymlValidator(val) == False:
                    return False
            return True
        elif type(node) == sigma.parser.condition.ConditionNOT:
            if self.ymlValidator(node.items) ==  False:
                return False
            return True
        elif type(node) == sigma.parser.condition.ConditionNULLValue:
            return True
        elif type(node) == sigma.parser.condition.ConditionNotNULLValue:
            return True
        elif type(node) == sigma.parser.condition.NodeSubexpression:
            if self.ymlValidator(node.items) == False:
                return False
            return True
        elif type(node) == tuple:
            fieldname, value = node
            if fieldname is None:
                return False
            return True
        elif type(node) in (str, int):
            return False
        elif type(node) == list:
            for value in node:
                if self.ymlValidator(value) == False:
                    return False
            return True
        else:
            return False
        

    def notSupportedLogsource(self, product, service):
        if product is None:
            product = ""

        if service is None:
            service = ""

        for key, val in self.fileFilterDicts.items():
            if key == product:
                if val == "":
                    return True
                else:
                    val = ",%s," % val
                    service = ",%s," % service
                    if val.find(service) != -1:
                        return True
        return False

    def loadCSVfiles(self):
        #It's used to map field name to internal attributes in FSIM
        if len(self.ymlAttr2FortiSIEMAttr) == 0:
            with open("./tools/config/fortisiem/FortiSIEM_EventAttributeMapping.csv", newline='') as csvfile:
                spamreader = csv.reader(csvfile, delimiter=',')
                for row in spamreader:
                    if len(row) < 2:
                        continue;
                    elif len(row) == 2:
                        self.fortiSIEMAttrType[row[1]] = "string"
                    else:
                        self.fortiSIEMAttrType[row[1]] = row[2]

                    self.ymlAttr2FortiSIEMAttr[row[0]] = row[1]
        #It's used to map event id to event type.
        if len(self.WindowsSysmonCode2FortiSIEMEvtTy) == 0:
            with open("./tools/config/fortisiem/FortiSIEM_SysMonEventTypeMapping.csv", newline='') as csvfile:
                spamreader = csv.reader(csvfile, delimiter=',')
                for row in spamreader:
                    if len(row) > 1:
                        self.WindowsSysmonCode2FortiSIEMEvtTy[row[0]] = row[1]

        #It's used to skip some files. When yml file match the constraints in it, we don't need generate rule from that yml file.
        if len(self.fileFilterDicts) == 0:
             with open("./tools/config/fortisiem/FortiSIEM_SkipUnsupportedLogSources.csv", newline='') as csvfile:
                 spamreader = csv.reader(csvfile, delimiter=',')
                 for row in spamreader:
                     if len(row) > 1:
                         self.fileFilterDicts[row[0]] = row[1]


    def loadMitreAttackMatrixFile(self, backend_options):
        techniquefile = backend_options.get("attackMapFile", None)
        if techniquefile is None:
            return

        if len(self.techniqueMap) == 0:
            with open(techniquefile, newline='') as f:
                spamreader = csv.reader(f, delimiter=',')
                for row in spamreader:
                    if len(row) < 3:
                        continue
                    else:
                        if row[2] != "":
                            self.techniqueMap[row[0]] = row[2]

    def formatSubFunctionAndTechniqueId(self, techniqueIds):
        sub_function_str = "Persistence"; 
        technique_str = None
        techniqueIds = sorted(techniqueIds)
        if len(techniqueIds) == 0:
           return sub_function_str, technique_str

        technique_str = ','.join(techniqueIds)
        for item in techniqueIds:
            tmp = self.techniqueMap.get(item, None)
            if tmp is None:
                continue
            tmp = tmp.split(",")
            sub_function_str = tmp[0];
            break;
        
        return sub_function_str, technique_str

    def formatRuleName(self, name):
        #ruleName has invalid characters. It only accepts: a-zA-Z0-9 \/:.$-
        ruleName = re.sub('\s*[^a-zA-Z0-9 \/:.$_\'\"-]+\s*', ' ', name)
        ruleName = re.sub('_', '-', ruleName)
        ruleName = re.sub('[\'"\(\)+,]*', '', ruleName)
        return ruleName

    def formatRuleTitle(self, name):
        #IncidentTitle has invalid characters. It only accepts: a-zA-Z0-9 _$-
        titleName = re.sub('\s*[^a-zA-Z0-9 _-]+\s*', ' ', name)
        return titleName;

    def setRuleType(self, backend_options):
        ruletype = backend_options.get("ruleType", None)
        ruleStartIndex = backend_options.get("ruleIndex", None)
        if ruletype is not None:
            self.ruleType = ruletype
        if ruleStartIndex is not None:
            self.ruleIndex = int(ruleStartIndex)

    def convertFieldNameToInterAttrName(self, fieldname):
        val = self.ymlAttr2FortiSIEMAttr.get(fieldname, None)
        if val is None:
            interfieldname = fieldname
        else:
            interfieldname = val

        self.curAttrs.add(interfieldname)
        return interfieldname

    def convertFieldValToInterVal(self, fieldname, value):
        val = self.generateValueNode(value);

        interfieldname = self.convertFieldNameToInterAttrName(fieldname)

        attrType = self.fortiSIEMAttrType.get(interfieldname, None)
        if val.find('.*') != -1:
             attrType = "string";

        if interfieldname == "eventType" and val.isdigit():
            val = self.formatEvtTypeVal(val)
        elif attrType is None or attrType == "string":
            val = self.strValueExpression % val;

        return val

    def formatEvtTypeVal(self, code):
        val = "\".*%s.*\"" % code
        if self.product == "windows":
            val = "\"Win-.*%s.*\"" % code

        if self.product == "windows" and ( self.service == "sysmon" or (self.service is None and  self.category == "sysmon")) :
            val = self.WindowsSysmonCode2FortiSIEMEvtTy.get(code, None)
            if val is None:
                val = "\"Win-Sysmon-%s-.*\"" % code
            else:
                evt = val.split(",")
                val = ",".join(["\"%s\"" % item for item in evt])
        elif self.product == "windows" and (self.service == "system" or (self.service is None and  self.category == "system")):
            val = "\"Win-System-%s\"" % code
        elif self.product == "windows" and ( ( self.service == "powershell" or self.service == "powershell-classic") or (self.service is None and ( self.category == "powershell" or self.category == "powershell-classic"))):
            val = "\"Win-PowerShell-%s\"" % code
        elif self.product == "windows" and ( self.service == "security" or (self.service is None and  self.category == "security")):
            val = "\"Win-Security-%s\"" % code
        elif self.product == "windows" and ( self.service == "application" or (self.service is None and  self.category == "application")):
            if self.sourceValueForWinAppEvtTy is not None:
                val = "\"Win-App-%s-%s\"" % ( self.sourceValueForWinAppEvtTy, code)
        return val

    def convertStrToRegstr(self, value):
        val = value.replace('\\"', '"')
        val = val.replace('\\', "\\\\")
        val = val.replace('"', '\\"')
        val = val.replace('.', "\\\\.")
        val = val.replace('(', "\\\\(")
        val = val.replace(')', "\\\\)")
        val = val.replace('&', "&amp;")
        val = val.replace('<', "&lt;")
        val = val.replace('>', "&gt;")
        val = val.replace('[', "\\\\[")
        val = val.replace(']', "\\\\]")
        val = val.replace('|', "\\\\|")
        val = val.replace('}', "\\\\}")
        val = val.replace('{', "\\\\{")
        val = val.replace('^', "\\\\^")
        val = val.replace('$', "\\\\$")
        val = val.replace('+', "\\\\+")
        val = val.replace('!', "\\\\!")
        val = val.replace('?', "\\\\?")
        val = val.replace('*', ".*")
        return val 

    def convertStrToXmlstr(self, value):
        val = value.replace('&', "&amp;")
        return val

    def generateQuery(self, parsed):
        result = self.generateNode(parsed.parsedSearch)
        if type(parsed.parsedSearch) == sigma.parser.condition.NodeSubexpression:
           result =  result[1:-1]   
        return result

    def generateMapItemNode(self, node, regdicts, isnot=False):
        fieldname, value = node
        interName = self.convertFieldNameToInterAttrName(fieldname)

        mapExp = self.mapExpression
        regExp = self.regExpression
        nullExp = self.nullExpression
        listExp = self.mapListValueExpression
        if isnot:
            mapExp = self.mapExpressionNot
            regExp = self.regExpressionNot
            nullExp = self.notNullExpression
            listExp = self.mapListValueExpressionNot

        if not value:
            return nullExp % (interName, )
        elif self.mapListsSpecialHandling == False and type(value) in (str, int, list) or self.mapListsSpecialHandling == True and type(value) in (str, int):
            val = self.convertFieldValToInterVal(interName, value)
            if len(val) > self.MAX_LEN:
                self.isValTooLong = True

            if interName== "eventType":
                if val.find(".*") == -1 and val.find(",") == -1:
                    return mapExp % (interName, val)
                elif val.find(".*") == -1 and val.find(",") != -1:
                    return listExp % (interName, val) 
                else:
                    return regExp % (interName, val)
            else:
                if val.find(".*") == -1:
                    return mapExp % (interName, val)
                else:
                    return regExp % (interName, val)

        elif type(value) == list:
            #print(self.generateMapItemListNode(interName, value, isnot))
            return self.generateMapItemListNode(interName, value, isnot)
        elif isinstance(value, SigmaTypeModifier):
            val = regdicts.get(interName, None)
            if val is None: 
                regList = set()
                regList.add(value.__str__())
                regdicts[interName] = regList
            else:
                val.add(value.__str__())
                regdicts[interName] = val
            return None;
            #return self.generateMapItemTypedNode(interName, value)
        else:
            raise TypeError("Backend does not support map values of type " + str(type(value)))

    def generateRegExpresion(self, regD, isnot=False):
        if len(regD) == 0:
            return None

        result = ""
        for key in regD.keys():
            vals = regD[key]
            res = ""
            if type(vals) == set:
                vals = sorted(vals)
                for val in vals:
                    res = res + ("|%s" % val)
                res = res[1:]
                res = "\"%s\"" % res
               
            if result != "" and isnot:
                result += " AND "
            elif result != "" and isnot==False:
                result += " OR "

            if len(res) > self.MAX_LEN:
                self.isValTooLong = True

            if isnot:
                result += self.regExpressionNot % (key, res)
            else:
                result += self.regExpression % (key, res)
            result = result.replace('\\', "\\\\")
        #print(result)
        return result

    def generateMapItemListNode(self, key, value, isnot=False):
        key = self.convertFieldNameToInterAttrName(key)

        if not set([type(val) for val in value]).issubset({str, int}):
            raise TypeError("List values must be strings or numbers")
        tmp = []
        tmpReg = []
        for item in value:
            val = self.convertFieldValToInterVal(key, item)
            valset = set()
            if key == "eventType": 
                valset = val.split(",")
            else:
                valset.add(val)

            for item in valset:
                if item.find(".*") == -1:
                    tmp.append(item)
                else:
                    tmpReg.append(val)        

        tmpstr=''
        mapExp = self.mapExpression
        mapListExp = self.mapListValueExpression
        regExp = self.regExpression
        if isnot:
            mapExp = self.mapExpressionNot
            mapListExp = self.mapListValueExpressionNot
            regExp = self.regExpressionNot

        if len(tmp) == 1:
            tmpstr = (mapExp %  (key, tmp[0]))
        elif len(tmp) > 1:
            tmp = sorted(tmp)
            tmpstr = (",".join(['%s' % (item) for item in tmp]))
            tmpstr = mapListExp % (key, tmpstr)
        
        tmpregstr=''
        if len(tmpReg) > 0:
            tmpReg = sorted(tmpReg)
            tmpregstr = ('|'.join(tmpReg))
            tmpregstr = tmpregstr.replace('"|"', '|')
            if len(tmpregstr) > self.MAX_LEN:
                self.isValTooLong = True

            tmpregstr = regExp % (key, tmpregstr)

        if tmpstr != '' and tmpregstr != '' and isnot:
          return "( %s AND %s )" % (tmpstr, tmpregstr)
        elif tmpstr != '' and tmpregstr != '':
          return "( %s OR %s )" % (tmpstr, tmpregstr)
        elif tmpstr == '':
            return tmpregstr
        else:
            return tmpstr

    def generateNode(self, node, regdicts={}):
        if type(node) == sigma.parser.condition.ConditionAND:
            return self.applyOverrides(self.generateANDNode(node))
        elif type(node) == sigma.parser.condition.ConditionOR:
            return self.applyOverrides(self.generateORNode(node))
        elif type(node) == sigma.parser.condition.ConditionNOT:
            return self.applyOverrides(self.generateNOTNode(node))
        elif type(node) == sigma.parser.condition.ConditionNULLValue:
            return self.applyOverrides(self.generateNULLValueNode(node))
        elif type(node) == sigma.parser.condition.ConditionNotNULLValue:
            return self.applyOverrides(self.generateNotNULLValueNode(node))
        elif type(node) == sigma.parser.condition.NodeSubexpression:
            return self.applyOverrides(self.generateSubexpressionNode(node))
        elif type(node) == tuple:
            return self.applyOverrides(self.generateMapItemNode(node, regdicts))
        elif type(node) in (str, int):
            return self.applyOverrides(self.generateValueNode(node))
        elif type(node) == list:
            return self.applyOverrides(self.generateListNode(node))
        elif isinstance(node, SigmaTypeModifier):
            return self.applyOverrides(self.generateTypedValueNode(node))
        else:
            raise TypeError("Node type %s was not expected in Sigma parse tree" % (str(type(node))))
        
   #A AND NOT B ---> != 
    def covertToNotValue(self, item, regdicts={}):
        if type(item) == sigma.parser.condition.ConditionAND:
            return self.applyOverrides(self.convertANDToORNode(item))
        elif type(item) == sigma.parser.condition.ConditionOR:
            return self.applyOverrides(self.covertORToANDNode(item))
        elif type(item) == sigma.parser.condition.ConditionNOT:
            return self.applyOverrides(self.generateNode(item))
        elif type(item) == sigma.parser.condition.ConditionNULLValue:
            return self.applyOverrides(self.generateNotNULLValueNode(item))
        elif type(item) == sigma.parser.condition.ConditionNotNULLValue:
            return self.applyOverrides(self.generateNULLValueNode(item))
        elif type(item) == sigma.parser.condition.NodeSubexpression:
            return self.applyOverrides(self.generateNotSubexpressionNode(item))
        elif type(item) == tuple:
            return self.applyOverrides(self.generateMapItemNode(item, regdicts, isnot=True))
        elif type(item) in (str, int):
            return self.applyOverrides(self.generateValueNode(item))
        elif type(item) == list:
            return self.applyOverrides(self.generateListNode(item))
        elif isinstance(item, SigmaTypeModifier):
            return self.applyOverrides(self.generateTypedValueNode(item))
        elif type(item) == sigma.parser.condition.ConditionNOT:
            return self.applyOverrides(self.generateNode(item.item))
        else:
            raise TypeError("Node type %s was not expected in Sigma parse tree" % (str(type(item))))

    def generateANDNode(self, node):
        generated = [ self.generateNode(val) for val in node ]
        filtered = [ g for g in generated if g is not None ]
        if filtered:
            if self.sort_condition_lists:
                filtered = sorted(filtered)
            return self.andToken.join(filtered)
        else:
            return None

    def generateORNode(self, node):
        regDicts = {}
        generated = [ self.generateNode(val, regDicts) for val in node ]

        res = None
        filtered = [ g for g in generated if g is not None ]
        if filtered:
            if self.sort_condition_lists:
                filtered = sorted(filtered)
            res = self.orToken.join(filtered)

        if len(regDicts) == 0:
            return res
        else:
            tmp = self.generateRegExpresion(regDicts)
            regDicts.clear()
            if res is None:
                return tmp
            else:
                return res + " OR " + tmp

    def convertANDToORNode(self, node):
        regDicts = {}
        generated = [ self.covertToNotValue(val, regDicts) for val in node ]

        res = None
        filtered = [ g for g in generated if g is not None ]
        if filtered:
            if self.sort_condition_lists:
                filtered = sorted(filtered)
            res = self.orToken.join(filtered)

        if len(regDicts) == 0:
            return res
        else:
            tmp = self.generateRegExpresion(regDicts, isnot=True)
            regDicts.clear()
            if res is None:
                return tmp
            else:
                return res + " OR " + tmp

    def covertORToANDNode(self, node):
        generated = [ self.covertToNotValue(val) for val in node ]
        filtered = [ g for g in generated if g is not None ]
        if filtered:
            if self.sort_condition_lists:
                filtered = sorted(filtered)
            return self.andToken.join(filtered)
        else:
            return None

    def generateNOTNode(self, node):
        item = node.item
        generated = self.covertToNotValue(item)
        return generated

    def generateNotSubexpressionNode(self, node):
        generated = self.covertToNotValue(node.items)
        if generated:
            return self.subExpression % generated
        else:
            return None

    def generateSubexpressionNode(self, node):
        generated = self.generateNode(node.items)
        if generated:
            return self.subExpression % generated
        else:
            return None

    def generateValueNode(self, node):
        val = self.cleanValue(str(node))
        if val.find('*') != -1:
            val = self.convertStrToRegstr(val)
        else:
            val = self.convertStrToXmlstr(val)

        return str(self.valueExpression % (val))

    def generateNULLValueNode(self, node):
        interName = self.convertFieldNameToInterAttrName(node.item)
        return self.nullExpression % (interName)

    def generateNotNULLValueNode(self, node):
        interName = self.convertFieldNameToInterAttrName(node.item)
        return self.notNullExpression % (interName)

    def generateRuleHeader(self):
        rulename = "PH_SYS_RULE_THREAT_HUNTING"
        if self.ruleType is None:
            ruleId = "PH_Rule_SIGMA_%d" % (self.ruleIndex)
        else:
            ruleId = "PH_Rule_%s_SIGMA_%d" % (self.ruleType, self.ruleIndex)
 
        f = open(self.ymlFileName, 'r') 
        lines = f.readlines() 
        tags = set()

        for line in lines: 
            match = re.search('^\s*-\s*attack.', line)
            if match is None:
                continue

            index = line.find("an old one")
            if index != -1:
                continue
            line = re.sub("\s*\\n$", '', line)

            tags.add(line)
        f.close()

        technique = []
        for tag in tags:
            tag = re.sub('^\s*-\s*attack\.\s*','', tag)
            match = re.search('(t|T)(\d+\.\d+|\d+)\s*', tag)
            if match is not None:
                tag = tag[1:]
                technique.append("T%s" % tag)
            else:
                match = re.search('\d', tag)
                if match is not None:
                    continue
                tag = re.sub('_',' ', tag).title()
        
        sub_function_str, technique_str= self.formatSubFunctionAndTechniqueId(technique)

        result = None
        if technique_str is not None:
            result = ("<Rule group=\"%s\" natural_id=\"%s\"  phIncidentCategory=\"Server\" function=\"Security\" subFunction=\"%s\" technique=\"%s\">") % (rulename, ruleId, sub_function_str, technique_str)
        else:
            result = ("<Rule group=\"%s\" natural_id=\"%s\"  phIncidentCategory=\"Server\" function=\"Security\" subFunction=\"%s\">") % (rulename, ruleId, sub_function_str)

        return result,ruleId,technique_str

    def generateRuleCommonPart(self, name, description):
        curRuleName = self.formatRuleName(name)
        curTitleName = self.formatRuleTitle(name)

        description = self.convertStrToXmlstr(description)
        filestr = "";
        if self.ymlFileName is not None:
            filestr = "\n  <SigmaFileName> %s </SigmaFileName>" % self.ymlFileName

        tmp = ("\n  <Name>%s </Name>\n  <IncidentTitle>%s</IncidentTitle>\n  <active>true</active>\n  <Description> %s </Description>%s\n  <CustomerScope groupByEachCustomer=\"true\">\n     <Include all=\"true\"/>\n    <Exclude/>\n  </CustomerScope>") % (curRuleName, curTitleName, description, filestr)
        return tmp,curRuleName


    def generateRuleIncidentDef(self, name, level, attrset):
        if level == "low":
            severity = 3
        elif level == "medium":
            severity = 5
        elif level == "high":
            severity = 7
        elif level == "critical":
            severity = 9
        else:
            severity = 1

        title = self.convertStrToXmlstr(name)
        title = title.replace(" ", "_")
        ruleEvtType="PH_RULE_%s" % title
        
        filterStr = set()
        for item in attrset:
            if item == 'eventType':
                filterStr.add('compEventType = Filter.eventType')
            else:
                filterStr.add('%s = Filter.%s' % (item, item))

        filterStr=sorted(filterStr)
        arglist = ",".join(filterStr)
        curFilterAttrs = ",".join(attrset)

        result = ("\n  <IncidentDef eventType=\"%s\" severity=\"%d\">\n    <ArgList> %s </ArgList>\n  </IncidentDef>") % (ruleEvtType, severity, arglist) 
        return result,curFilterAttrs,ruleEvtType
         
    def generateRulePatternClause(self, evtConstrSet, groupByAttrs):
        singleEvtConstr = None
       
        if len(evtConstrSet) > 1:
            evtConstrSet = sorted(evtConstrSet)
            singleEvtConstr = (" OR ".join(['(%s)' % item for item in evtConstrSet]))
        else:
            singleEvtConstr = evtConstrSet.pop() 

        groupByAttr = ",".join(groupByAttrs)

        result = ("\n  <PatternClause window=\"300\">\n    <SubPattern displayName=\"Filter\" name=\"Filter\">\n    <SingleEvtConstr> %s </SingleEvtConstr>\n    <GroupByAttr> %s </GroupByAttr>\n    <GroupEvtConstr> COUNT(*) &gt;= 1 </GroupEvtConstr>\n    </SubPattern>\n    </PatternClause>") % (singleEvtConstr, groupByAttr)

        return result,groupByAttr

    def generateRuleTriggerEventDisplay(self, displayAttrs):
        displayAttrs = sorted(displayAttrs)
        if len(displayAttrs) == 0:
            fields = "phRecvTime,hostName,rawEventMsg"
            return ("\n  <TriggerEventDisplay>\n    <AttrList> %s </AttrList>\n  </TriggerEventDisplay>") % (fields)
        else:
            fields = "phRecvTime,hostName," +  ",".join(displayAttrs) + ",rawEventMsg"
            return ("\n  <TriggerEventDisplay>\n    <AttrList> %s </AttrList>\n  </TriggerEventDisplay>") % (fields)

    def generateRuleTailer(self):
        return "\n</Rule>\n"

    def getDisplayAttr(self, attrset):
        attrset.discard("hostName")
        attrset.discard("eventType")
        attrset.discard("phRecvTime")
        attrset = sorted(attrset)
        return attrset;
    
    def getIncidentDefAttr(self, attrset):
        attrset.add("hostName")
        attrset = sorted(attrset)
        return attrset;

    def getGroupByAttr(self, attrset):    
        attrset.add("hostName")
        attrset = sorted(attrset)
        return attrset

    def toCsvStr(self, val):
        if val is not None and  val.find(",") != -1:
            val = "\"%s\"" % val
        return val

    def generate(self, sigmaparser):

        result = set()

        date = sigmaparser.parsedyaml["date"]
        name = sigmaparser.parsedyaml["title"]    
        des = sigmaparser.parsedyaml["description"]
        level = sigmaparser.parsedyaml["level"]

        res,errMsg = self.generateEvtConstrForOneLogsource(sigmaparser);
        if errMsg is not None:
            print("%s, %s" % (self.ymlFileName, errMsg))
            return None

        result.add(res)
        groupByAttr=copy.deepcopy(self.curAttrs)
        displayAttr=copy.deepcopy(self.curAttrs)
        incidentDefAttr=copy.deepcopy(self.curAttrs)

        groupByAttr = self.getGroupByAttr(groupByAttr)
        displayAttr = self.getDisplayAttr(displayAttr)
        incidentDefAttr = self.getIncidentDefAttr(incidentDefAttr)

        ruleHeader,ruleId,techniques = self.generateRuleHeader()
        ruleCommonPart,curRuleName = self.generateRuleCommonPart(name, des)
        ruleIncidentDef,filterAttrStr,ruleEvtType = self.generateRuleIncidentDef(name, level, incidentDefAttr)
        ruleTriggerEventDisplay = self.generateRuleTriggerEventDisplay(displayAttr)
        ruleTailer = self.generateRuleTailer()
        rulePatternClause,groupByStr = self.generateRulePatternClause(result, groupByAttr)

        result = ruleHeader + ruleCommonPart + ruleIncidentDef + rulePatternClause + ruleTriggerEventDisplay + ruleTailer
        self.ruleIndex += 1
        return result
    
    def setSourceValueForWinServiceApp(self, sigmaparser):
        detection = sigmaparser.parsedyaml.get("detection", None)
        svtSource = set()
        for val in detection.values():
            if isinstance(val,dict):
                for k, v in val.items():
                    if k == "Source":
                        svtSource.add(v)
        svtSource = sorted(svtSource)
        if len(svtSource) == 1:
            self.sourceValueForWinAppEvtTy = svtSource[0]
        else:
            self.sourceValueForWinAppEvtTy = None

    def generateEvtConstrForOneLogsource(self, sigmaparser):
        errMsg = None
        result = None
        self.curAttrs = set()
        self.product = None
        self.service = None
        self.category = None
        self.isValTooLong = False
        logsource = sigmaparser.parsedyaml.get("logsource", None)
        if logsource is not None:
            self.product = logsource.get("product", None)
            self.service = logsource.get("service", None)
            self.category = logsource.get("category", None)

        if self.product == "windows" and ( self.service == "application" or (self.service is None and  self.category == "application")):
            self.setSourceValueForWinServiceApp(sigmaparser) 

        res = set()
        for parsed in sigmaparser.condparsed:
            if self.ymlFileName.find("deprecated/") != -1:
               errMsg = "SKIP RULE (YML is in deprecated directory)"
            elif self.notSupportedLogsource(self.product, self.service):
               errMsg = "SKIP RULE (Logsource in SkipRuleOflogsource.csv)"
            elif self.ymlValidator(parsed.parsedSearch) == False:
                errMsg = "SKIP RULE(Yml format is wrong)"
            if parsed.parsedAgg:
                errMsg = "SKIP RULE (There is an aggregation operator not implemented)"
            else: # generate Event Constr 
                result = self.generateQuery(parsed)
                if result is None:
                    errMsg = "SKIP RULE (There is no single Event Constraint)"
                elif self.isValTooLong:
                    errMsg = "SKIP RULE (Regular expression is too long)"
                else:
                    if self.product == "windows":
                        if result.find('eventType') == -1:
                            errMsg = "SKIP RULE (There is no event type in constraint)"
                            return None, errMsg

                    res.add(result)

        if len(res) == 1:
            result = res.pop()
        else:
            result = " OR ".join(["(%s)"%item for item in res])

        return result,errMsg



# Azure Log Analytics output backend for sigmac
# John Tuckner (@tuckner)

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
import sys
import re
import json
import xml.etree.ElementTree as xml

from datetime import timedelta
from uuid import uuid4

from sigma.config.mapping import (
    SimpleFieldMapping, MultiFieldMapping, ConditionalFieldMapping
)
from sigma.parser.condition import SigmaAggregationParser, SigmaConditionParser, SigmaConditionTokenizer

from sigma.parser.modifiers.type import SigmaRegularExpressionModifier
from sigma.backends.base import SingleTextQueryBackend

from sigma.parser.modifiers.base import SigmaTypeModifier
from sigma.parser.modifiers.transform import SigmaContainsModifier, SigmaStartswithModifier, SigmaEndswithModifier
from .data import sysmon_schema
from .exceptions import NotSupportedError

class AzureLogAnalyticsBackend(SingleTextQueryBackend):
    """Converts Sigma rule into Azure Log Analytics Queries."""
    identifier = "ala"
    active = True
    options = SingleTextQueryBackend.options + (
        ("sysmon", False, "Generate Sysmon event queries for generic rules", None),
        (
            "use_fields",
            False,
            "Use fields to generate project and aggregation clauses",
            None,
        ),
    )
    config_required = False

    reEscape = re.compile('(\\\|"|(?<!)(?![*?]))')
    andToken = " and "
    orToken = " or "
    notToken = "not "
    subExpression = "(%s)"
    listExpression = "(%s)"
    listSeparator = ", "
    valueExpression = "\"%s\""
    nullExpression = "isnull(%s)"
    notNullExpression = "isnotnull(%s)"
    mapExpression = "%s == %s"
    mapListsSpecialHandling = True
    mapListValueExpression = "%s in %s"
    typedValueExpression = {
        SigmaRegularExpressionModifier: "matches regex \"(?i)%s\"",
        SigmaContainsModifier: "contains \"%s\""
    }

    # _WIN_SECURITY_EVENT_MAP = {
    #     "Image": "NewProcessName",
    #     "ParentImage": "ParentProcessName",
    #     "User": "SubjectUserName",
    # }

    def __init__(self, *args, **kwargs):
        """Initialize field mappings."""
        super().__init__(*args, **kwargs)
        self.category = None
        self.product = None
        self.service = None
        self.table = None
        self.eventid = None
        self.tableAggJoinFields = None
        self.tableAggTimeField = None
        self._parser = None
        self._fields = None
        self._agg_var = None
        self._has_logsource_event_cond = False
        if not self.sysmon and not self.sigmaconfig.config:
            self._field_map = {}#self._WIN_SECURITY_EVENT_MAP
        else:
            self._field_map = {}

    def map_sysmon_schema(self, eventid):
        schema_keys = []
        try:
            tree = xml.ElementTree(xml.fromstring(sysmon_schema))
        except:
            raise NotSupportedError("Required Sysmon schema not provided")
        root = tree.getroot()
        for child in root.iter('event'):
            if child.attrib['value'] == str(eventid):
                for d in list(child):
                    schema_keys.append(d.attrib["name"])
        parse_arg = ''
        for schema_key in schema_keys:
            parse_arg += "'Data Name=\"{0}\">' {0} '<' * ".format(schema_key)
        return parse_arg

    def default_value_mapping(self, val):
        if isinstance(val, int):
            return "== %d" % (val)
        op = "=="
        if isinstance(val, str):
            if "*" in val[1:-1]:  # value contains * inside string - use regex match
                op = "matches regex"
                val = re.sub('(\\\\\*|\*)', '.*', val)
                if "\\" in val:
                    val = "@'(?i)%s'" % (val)
                else:
                    val = "'(?i)%s'" % (val)
                return "%s %s" % (op, self.cleanValue(val))
            elif val.startswith("*") or val.endswith("*"):
                if val.startswith("*") and val.endswith("*"):
                    op = "contains"
                    val = val[1:-1]
                elif val.startswith("*"):
                    op = "endswith"
                    val = val[1:]
                elif val.endswith("*"):
                    op = "startswith"
                    val = val[:-1]
                val = re.sub('([".^$]|(?![*?]))', '\g<1>', val)
                val = re.sub('(\\\\\*|\*)', '', val)
                val = re.sub('\\?', '.', val)
                if "\\" in val:
                    return "%s @'%s'" % (op, self.cleanValue(val))
                return "%s '%s'" % (op, self.cleanValue(val))
            elif "\\" in val:
                return "%s @'%s'" % (op, self.cleanValue(val))
        return "%s \"%s\"" % (op, self.cleanValue(val))

    def getTable(self, sigmaparser):
        if self.category == "process_creation" and len(set(sigmaparser.values.keys()) - {"Image", "ParentImage",
                                                                                         "CommandLine"}) == 0:
            self.table = "SecurityEvent | where EventID == 4688 "
            self.eventid = "4688"
            self.tableAggJoinFields = "SubjectLogonId, Computer"
            self.tableAggTimeField = "TimeGenerated"
        elif self.category == "process_creation":
            self.table = "SysmonEvent"
            self.eventid = "1"
        elif self.service and self.service.lower() == "security":
            self.table = "SecurityEvent"
        elif self.service and self.service.lower() == "sysmon":
            self.table = "SysmonEvent"
        elif self.service and self.service.lower() == "powershell":
            self.table = "Event"
        elif self.service and self.service.lower() == "office365":
            self.table = "OfficeActivity"
        elif self.service and self.service.lower() == "azuread":
            self.table = "AuditLogs"
        elif self.service and (self.service.lower() == "azureactivity" or self.service.lower() == "activitylogs"):
            self.table = "AzureActivity"
        else:
            if self.service:
                if "-" in self.service:
                    self.table = "-".join([item.capitalize() for item in self.service.split("-")])
                elif "_" in self.service:
                    self.table = "_".join([item.capitalize() for item in self.service.split("_")])
                else:
                    if self.service.islower() or self.service.isupper():
                        self.table = self.service.capitalize()
                    else:
                        self.table = self.service
            elif self.product:
                if "-" in self.product:
                    self.table = "-".join([item.capitalize() for item in self.product.split("-")])
                elif "_" in self.product:
                    self.table = "_".join([item.capitalize() for item in self.product.split("_")])
                else:
                    if self.product.islower() or self.product.isupper():
                        self.table = self.product.capitalize()
                    else:
                        self.table = self.product
            elif self.category:
                if "-" in self.category:
                    self.table = "-".join([item.capitalize() for item in self.category.split("-")])
                elif "_" in self.category:
                    self.table = "_".join([item.capitalize() for item in self.category.split("_")])
                else:
                    if self.category.islower() or self.category.isupper():
                        self.table = self.category.capitalize()
                    else:
                        self.table = self.category

    def generate(self, sigmaparser):
        try:
            self.category = sigmaparser.parsedyaml['logsource'].setdefault('category', None)
            self.product = sigmaparser.parsedyaml['logsource'].setdefault('product', None)
            self.service = sigmaparser.parsedyaml['logsource'].setdefault('service', None)
        except KeyError:
            self.category = None
            self.product = None
            self.service = None
        detection = sigmaparser.parsedyaml.get("detection", {})
        if "keywords" in detection.keys():
            return super().generate(sigmaparser)
        if self.table is None:
            self.getTable(sigmaparser)

        return super().generate(sigmaparser)

    def generateBefore(self, parsed):
        if self.table is None:
            raise NotSupportedError("No table could be determined from Sigma rule")
        if self.category == "process_creation" and self.sysmon:
            parse_string = self.map_sysmon_schema(self.eventid)
            before = "%s | parse EventData with * %s | where EventID == \"%s\" | where " % (self.table, parse_string, self.eventid)
        elif self.sysmon:
            parse_string = self.map_sysmon_schema(self.eventid)
            before = "%s | parse EventData with * %s | where " % (self.table, parse_string)
        # elif self.category == "process_creation" and not self._has_logsource_event_cond:
        #     before = "%s | where EventID == \"%s\" | where " % (self.table, self.eventid)
        else:
            before = "%s | where " % self.table
        if parsed.parsedAgg != None and parsed.parsedAgg.aggfunc == SigmaAggregationParser.AGGFUNC_NEAR:
            window = parsed.parsedAgg.parser.parsedyaml["detection"].get("timeframe", "30m")
            before = """
            let lookupWindow = %s;
            let lookupBin = lookupWindow / 2.0;
            """ % (window) + before
        return before

    def generateMapItemNode(self, node):
        """
        ALA queries, like ATP, refer to event tables instead of Windows logging event identifiers. This method catches conditions that refer to this field
        and creates an appropriate table reference.
        """
        key, value = node
        key = self.fieldNameMapping(key, value)
        if type(value) == list:         # handle map items with values list like multiple OR-chained conditions
            return "(" + self.generateORNode(
                    [(key, v) for v in value]
                    ) + ")"
        elif key == "EventID":            # EventIDs are not reflected in condition but in table selection
            if self.service == "sysmon":
                self.table = "SysmonEvent"
                self.eventid = value
            elif self.service == "powershell":
                self.table = "Event"
            elif self.service == "security":
                self.table = "SecurityEvent"
            elif self.service == "system":
                self.table = "Event"
            return self.mapExpression % (key, value)
        elif type(value) in [SigmaTypeModifier, SigmaContainsModifier, SigmaRegularExpressionModifier, SigmaStartswithModifier, SigmaEndswithModifier]:
            return self.generateMapItemTypedNode(key, value)
        elif type(value) in (str, int):    # default value processing'
            #default_filters = ["endswith", "contains", "startswith", "re"]
            # if any([item for item in default_filters if item in key]):
            #     key = re.sub(key, default_filters, "")
            #     return self.regexExpression % (key, self.cleanValue(value))
            # else:
            #     value_mapping = self.default_value_mapping
            value_mapping = self.default_value_mapping
            mapping = (key, value_mapping)
            if len(mapping) == 1:
                mapping = mapping[0]
                if type(mapping) == str:
                    return mapping
                elif callable(mapping):
                    return self.generateSubexpressionNode(
                            self.generateANDNode(
                                [cond for cond in mapping(key, self.cleanValue(value))]
                                )
                            )
            elif len(mapping) == 2:
                result = list()
                for mapitem, val in zip(mapping, node):     # iterate mapping and mapping source value synchronously over key and value
                    if type(mapitem) == str:
                        result.append(mapitem)
                    elif callable(mapitem):
                        result.append(mapitem(self.cleanValue(val)))
                return "{} {}".format(*result)
            else:
                raise TypeError("Backend does not support map values of type " + str(type(value)))
        elif type(value) == list:
            return self.generateMapItemListNode(key, value)

        elif value is None:
            return self.nullExpression % (key, )
        else:
            raise TypeError("Backend does not support map values of type " + str(type(value)))

    def generateMapItemTypedNode(self, fieldname, value):
        return "%s %s" % (fieldname, self.generateTypedValueNode(value))

    def generateTypedValueNode(self, node):
        try:
            val = str(node)
            if "*" in val:
                val = re.sub('\\*', '.*', val)
            return self.typedValueExpression[type(node)] % (val)
        except KeyError:
            raise NotImplementedError("Type modifier '{}' is not supported by backend".format(node.identifier))

    def generateAggregationQuery(self, agg, searchId):
        condtoken = SigmaConditionTokenizer(searchId)
        condparsed = SigmaConditionParser(agg.parser, condtoken)
        backend = AzureLogAnalyticsBackend(agg.config)

        # these bits from generate() should be moved to __init__
        try:
            backend.category = agg.parser.parsedyaml['logsource'].setdefault('category', None)
            backend.product = agg.parser.parsedyaml['logsource'].setdefault('product', None)
            backend.service = agg.parser.parsedyaml['logsource'].setdefault('service', None)
        except KeyError:
            backend.category = None
            backend.product = None
            backend.service = None
        backend.getTable(agg.parser)

        query = backend.generateQuery(condparsed)
        before = backend.generateBefore(condparsed)
        return before + query

    # follow the join/time window pattern
    # https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/join-timewindow
    def generateNear(self, agg):
        includeQueries = []

        includeCount = 0
        for includeCount, include in enumerate(agg.include, start=1):
            iq = self.generateAggregationQuery(agg, include)
            iq += """
                | extend End{timeIndex}={timeField},
                    TimeKey = range(
                        bin({timeField} - lookupWindow, lookupBin),
                        bin({timeField}, lookupBin),
                        lookupBin)
                | mv-expand TimeKey to typeof(datetime)""".format(
                    timeField=self.tableAggTimeField,
                    timeIndex=includeCount,
                )
            includeQueries.append(iq)

        ret = " | extend Start={timeField}, TimeKey = bin({timeField}, lookupBin) | join kind=inner (\n  ".format(
            timeField=self.tableAggTimeField,
        )
        ret += ") on {joinFields}, TimeKey | join kind=inner (\n  ".format(
            joinFields=self.tableAggJoinFields,
        ).join(includeQueries)
        ret += ") on {joinFields}, TimeKey\n| where ".format(
            joinFields=self.tableAggJoinFields,
        )
        ret += " and ".join([
            "(End%d - Start) between (0min .. lookupWindow)" % (endIndex + 1) for endIndex in range(includeCount)
        ])

        return ret

    def generateAggregation(self, agg):
        if agg is None:
            return ""
        if agg.aggfunc == SigmaAggregationParser.AGGFUNC_NEAR:
            if agg.exclude:
                raise NotSupportedError("This backend doesn't currently support 'near' with excludes")
            if self.tableAggJoinFields == None or self.tableAggTimeField == None:
                raise NotSupportedError("This backend doesn't currently support 'near' for this table")
            return self.generateNear(agg)
        if agg.aggfunc_notrans != 'count' and agg.aggfield is None:
            raise NotSupportedError(
                "The '%s' aggregation operator " % agg.aggfunc_notrans
                + "must have an aggregation field for the %s backend" % self.identifier
            )
        if agg.aggfunc_notrans == 'count' and agg.aggfield is not None:
            agg_func = "dcount"
        else:
            agg_func = agg.aggfunc_notrans

        self._agg_var = agg_func + ("_" + agg.aggfield) if agg.aggfield else "var"
        if not self._fields:
            by_clause = "by {grp}".format(grp=agg.groupfield if agg.groupfield else "")
        else:
            if agg.aggfield_notrans in self._fields:
                self._fields.remove(agg.aggfield_notrans)
            by_clause = "by {grp}".format(grp=", ".join(self._map_fields(self._fields)))
        return (
            " | summarize {var} = {func}({fld}) {by} | where {var} {op} {cond}".format(
                var=self._agg_var,
                func=agg_func,
                fld=agg.aggfield or "",
                by=by_clause,
                op=agg.cond_op,
                cond=agg.condition,
            )
        )

    def _map_conditional_field(self, fieldname):
        mapping = self.sigmaconfig.fieldmappings.get(fieldname)
        # if there is a conditional mapping for this fieldname
        # and it matches the current event id, get the mapping
        if (
            mapping
            and isinstance(mapping, ConditionalFieldMapping)
            and "EventID" in mapping.conditions
        ):
            fieldname = mapping.conditions["EventID"].get(self.eventid, [fieldname])[0]
        elif self._field_map:
            # Fall back to default internal map if no config
            return self._field_map.get(fieldname, fieldname)
        return fieldname

class AzureAPIBackend(AzureLogAnalyticsBackend):
    """Converts Sigma rule into Azure Log Analytics Rule."""
    identifier = "ala-rule"
    active = True
    options = SingleTextQueryBackend.options + (
            ("sysmon", False, "Generate Sysmon event queries for generic rules", None),
            )

    def __init__(self, *args, **kwargs):
        """Initialize field mappings"""
        super().__init__(*args, **kwargs)
        self.techniques = self._load_mitre_file("techniques")

    def find_technique(self, key_ids):
        for key_id in set(key_ids):
            if not key_id:
                continue
            for technique in self.techniques:
                if key_id == technique.get("technique_id", ""):
                    yield technique
                if "." in key_id and key_id.split(".")[0] == technique.get("technique_id", ""):
                    yield technique

    def _load_mitre_file(self, mitre_type):
        try:
            backend_dir = os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "..", "config", "mitre"))
            path = os.path.join(backend_dir, "{}.json".format(mitre_type))
            with open(path) as config_file:
                config = json.load(config_file)
                return config
        except (IOError, OSError) as e:
            print("Failed to open {} configuration file '%s': %s".format(path, str(e)), file=sys.stderr)
            return []
        except json.JSONDecodeError as e:
            print("Failed to parse {} configuration file '%s' as valid YAML: %s" % (path, str(e)), file=sys.stderr)
            return []

    def skip_tactics_or_techniques(self, src_technics, src_tactics):
        tactics = set()
        technics = set()

        local_storage_techniques = {item["technique_id"]: item for item in self.find_technique(src_technics)}

        for key_id in src_technics:
            if "." in key_id:
                src_tactic = local_storage_techniques.get(key_id.split(".")[0], {}).get("tactic")
            else:
                src_tactic = local_storage_techniques.get(key_id, {}).get("tactic")
            if not src_tactic:
                continue
            src_tactic = set(src_tactic)

            for item in src_tactics:
                if item in src_tactic:
                    technics.add(key_id)
                    tactics.add(item)

        return sorted(tactics), sorted(technics)

    def parse_severity(self, old_severity):
        if old_severity.lower() == "critical":
            return "high"
        return old_severity

    def get_tactics_and_techniques(self, tags):
        tactics = list()
        technics = list()

        for tag in tags:
            tag = tag.replace("attack.", "")
            if re.match("[t][0-9]{4}", tag, re.IGNORECASE):
                technics.append(tag.title())
            else:
                if "_" in tag:
                    tag = tag.replace("_", " ")
                tag = tag.title()
                tactics.append(tag)

        return tactics, technics

    def timeframeToDelta(self, timeframe):
        time_unit = timeframe[-1:]
        duration = int(timeframe[:-1])
        return (
            time_unit == "s" and timedelta(seconds=duration) or
            time_unit == "m" and timedelta(minutes=duration) or
            time_unit == "h" and timedelta(hours=duration) or
            time_unit == "d" and timedelta(days=duration) or
            None
        )

    def iso8601_duration(self, delta):
        if not delta:
            return "PT0S"
        if not delta.seconds:
            return "P%dD" % (delta.days)
        days = delta.days and "%dD" % (delta.days) or ""
        hours = delta.seconds // 3600 % 24 and "%dH" % (delta.seconds // 3600 % 24) or ""
        minutes = delta.seconds // 60 % 60 and "%dM" % (delta.seconds // 60 % 60) or ""
        seconds = delta.seconds % 60 and "%dS" % (delta.seconds % 60) or ""
        return "P%sT%s%s%s" % (days, hours, minutes, seconds)

    def create_rule(self, config):
        tags = config.get("tags", [])

        tactics, technics = self.get_tactics_and_techniques(tags)
        tactics, technics = self.skip_tactics_or_techniques(technics, tactics)
        tactics = list(map(lambda s: s.replace(" ", ""), tactics))

        timeframe = self.timeframeToDelta(config["detection"].setdefault("timeframe", "30m"))
        queryDuration = self.iso8601_duration(timeframe)
        suppressionDuration = self.iso8601_duration(timeframe * 5)

        rule = {
                "displayName": "{} by {}".format(config.get("title"), config.get('author')),
                "description": "{} {}".format(config.get("description"), "Technique: {}.".format(",".join(technics))),
                "severity": self.parse_severity(config.get("level", "medium")),
                "enabled": True,
                "query": config.get("translation"),
                "queryFrequency": queryDuration,
                "queryPeriod": queryDuration,
                "triggerOperator": "GreaterThan",
                "triggerThreshold": 0,
                "suppressionDuration": suppressionDuration,
                "suppressionEnabled": True,
                "tactics": tactics
            }
        return json.dumps(rule)

    def generate(self, sigmaparser):
        translation = super().generate(sigmaparser)
        if translation:
            configs = sigmaparser.parsedyaml
            configs.update({"translation": translation})
            rule = self.create_rule(configs)
            return rule
        else:
            raise NotSupportedError("No table could be determined from Sigma rule")

class SentinelBackend(AzureAPIBackend):
    """Converts Sigma rule into Azure Sentinel scheduled alert rule ARM template."""
    identifier = "sentinel-rule"
    active = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def generate(self, sigmaparser):
        translation = super().generate(sigmaparser)
        if translation:
            configs = sigmaparser.parsedyaml
            configs.update({"translation": translation})
            rule = self.create_sentinel_rule(configs)
            return json.dumps(rule)

    def create_sentinel_rule(self, config):
        # https://docs.microsoft.com/en-us/azure/azure-resource-manager/templates/child-resource-name-type#outside-parent-resource
        # https://docs.microsoft.com/en-us/azure/templates/microsoft.operationalinsights/workspaces?tabs=json
        # https://docs.microsoft.com/en-us/rest/api/securityinsights/alert-rules/create-or-update#scheduledalertrule
        properties = json.loads(config.get("translation"))
        properties.update({
            "incidentConfiguration": {
                "createIncident": True,
                "groupingConfiguration": {
                    "enabled": False,
                    "reopenClosedIncident": False,
                    "lookbackDuration": properties['suppressionDuration'],
                    "matchingMethod": "AllEntities",
                    "groupByEntities": [],
                    "groupByAlertDetails": [],
                    "groupByCustomDetails": [],
                },
            },
            "eventGroupingSettings": {
                "aggregationKind": "SingleAlert",
            },
            "alertDetailsOverride": None,
            "customDetails": None,
            "templateVersion": "1.0.0",
        })
        rule_uuid = config.get("id", str(uuid4()))
        return {
            "$schema": "https://schema.management.azure.com/schemas/2019-04-01/deploymentTemplate.json#",
            "contentVersion": "1.0.0.0",
            "parameters": {
                "workspace": {
                    "type": "String",
                },
            },
            "resources": [
                {
                    "id": "[concat(resourceId('Microsoft.OperationalInsights/workspaces/providers', parameters('workspace'), 'Microsoft.SecurityInsights'),'/alertRules/" + rule_uuid + "')]",
                    "name": "[concat(parameters('workspace'),'/Microsoft.SecurityInsights/" + rule_uuid + "')]",
                    "type": "Microsoft.OperationalInsights/workspaces/providers/alertRules",
                    "apiVersion": "2021-03-01-preview",

                    "kind": "Scheduled",
                    "properties": properties,
                },
            ],
        }



# Output backends for sigmac
# Copyright 2016-2017 Thomas Patzke

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
from .base import BaseBackend
from .mixins import QuoteCharMixin

class GrepBackend(BaseBackend, QuoteCharMixin):
    """Generates Perl compatible regular expressions and puts 'grep -P -i' around it"""
    identifier = "grep"
    active = True
    config_required = False

    reEscape = re.compile("([\\|()\[\]{}.^$+])")

    def generateQuery(self, parsed):
        return "grep -P -i '^%s'" % self.generateNode(parsed.parsedSearch)

    def cleanValue(self, val):
        val = super().cleanValue(val)
        val = val.replace("'","'\"'\"'")
        return re.sub("\\*", ".*", val)

    def generateORNode(self, node):
        return "(?:%s)" % "|".join([".*" + self.generateNode(val) for val in node])

    def generateANDNode(self, node):
        return "".join(["(?=.*%s)" % self.generateNode(val) for val in node])

    def generateNOTNode(self, node):
        return "(?!.*%s)" % self.generateNode(node.item)

    def generateSubexpressionNode(self, node):
        return "(?:.*%s)" % self.generateNode(node.items)

    def generateListNode(self, node):
        if not set([type(value) for value in node]).issubset({str, int}):
            raise TypeError("List values must be strings or numbers")
        return self.generateORNode(node)

    def generateMapItemNode(self, node):
        key, value = node
        if value is None:
            return self.generateNULLValueNode(node)
        else:
            return self.generateNode(value)

    def generateValueNode(self, node):
        return self.cleanValue(str(node))

    def generateNULLValueNode(self, node):
        key, value = node
        return "(?!%s)" % key
